"""
Socket reader tests for pynmeagps - uses dummy socket class
to achieve 99% test coverage of SocketStream.

Created on 11 May 2022

*** NB: must be saved in UTF-8 format ***

:author: semuadmin (Steve Smith)
"""

import unittest
from socket import socket
from pynmeagps import NMEAReader, NMEAMessage, POLL


class DummySocket(socket):
    """
    Dummy socket class which simulates recv() method
    and TimeoutError.
    """

    def __init__(self, *args, **kwargs):
        self._timeout = False
        if "timeout" in kwargs:
            self._timeout = kwargs["timeout"]
            kwargs.pop("timeout")

        super().__init__(*args, **kwargs)

        pool = (
            b"\xb5b\x06\x8b\x0c\x00\x00\x00\x00\x00\x68\x00\x11\x40\xb6\xf3\x9d\x3f\xdb\x3d"
            + b"\xb5b\x10\x02\x1c\x00\x6d\xd8\x07\x00\x18\x20\x00\x00\xcd\x06\x00\x0e\xe4\xfe\xff\x0d\x03\xfa\xff\x05\x09\x0b\x00\x0c\x6d\xd8\x07\x00\xee\x51"
            + b"\xb5b\x10\x02\x18\x00\x72\xd8\x07\x00\x18\x18\x00\x00\x4b\xfd\xff\x10\x40\x02\x00\x11\x23\x28\x00\x12\x72\xd8\x07\x00\x03\x9c"
            + b"$GNDTM,W84,,0.0,N,0.0,E,0.0,W84*71\r\n"
            + b"$GNRMC,103607.00,A,5327.03942,N,10214.42462,W,0.046,,060321,,,A,V*0E\r\n"
            + b"$GPRTE,2,1,c,0,PBRCPK,PBRTO,PTELGR,PPLAND,PYAMBU,PPFAIR,PWARRN,PMORTL,PLISMR*73\r\n"
            + b"\xd3\x00\x13\x3e\xd7\xd3\x02\x02\x98\x0e\xde\xef\x34\xb4\xbd\x62\xac\x09\x41\x98\x6f\x33\x36\x0b\x98"
            + b"\xd3\x00\x13>\xd0\x00\x03\x8aX\xd9I<\x87/4\x10\x9d\x07\xd6\xafH Z\xd7\xf7"
            + b"\xd3\x00\x12B\x91\x81\xc9\x84\x00\x04B\xb8\x88\x008\x80\t\xd0F\x00(\xf0kf"
        )
        self._stream = pool * round(4096 / len(pool))
        self._buffer = self._stream

    def recv(self, num: int) -> bytes:
        if self._timeout:
            raise TimeoutError
        if len(self._buffer) < num:
            self._buffer = self._buffer + self._stream
        buff = self._buffer[:num]
        self._buffer = self._buffer[num:]
        return buff

    def send(self, data: bytes):
        if self._timeout:
            raise TimeoutError
        return None


class SocketTest(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def tearDown(self):
        pass

    # *******************************************
    # Helper methods
    # *******************************************

    def testSocketStub(self):
        EXPECTED_RESULTS = (
            "<NMEA(GNDTM, datum=W84, subDatum=, latOfset=0.0, NS=N, lonOfset=0.0, EW=E, alt=0.0, refDatum=W84)>",
            "<NMEA(GNRMC, time=10:36:07, status=A, lat=53.450657, NS=N, lon=-102.2404103333, EW=W, spd=0.046, cog=, date=2021-03-06, mv=, mvEW=, posMode=A, navStatus=V)>",
            "<NMEA(GPRTE, numMsg=2, msgNum=1, status=c, routeid=0, wpt_01=PBRCPK, wpt_02=PBRTO, wpt_03=PTELGR, wpt_04=PPLAND, wpt_05=PYAMBU, wpt_06=PPFAIR, wpt_07=PWARRN, wpt_08=PMORTL, wpt_09=PLISMR)>",
            "<NMEA(GNDTM, datum=W84, subDatum=, latOfset=0.0, NS=N, lonOfset=0.0, EW=E, alt=0.0, refDatum=W84)>",
            "<NMEA(GNRMC, time=10:36:07, status=A, lat=53.450657, NS=N, lon=-102.2404103333, EW=W, spd=0.046, cog=, date=2021-03-06, mv=, mvEW=, posMode=A, navStatus=V)>",
            "<NMEA(GPRTE, numMsg=2, msgNum=1, status=c, routeid=0, wpt_01=PBRCPK, wpt_02=PBRTO, wpt_03=PTELGR, wpt_04=PPLAND, wpt_05=PYAMBU, wpt_06=PPFAIR, wpt_07=PWARRN, wpt_08=PMORTL, wpt_09=PLISMR)>",
            "<NMEA(GNDTM, datum=W84, subDatum=, latOfset=0.0, NS=N, lonOfset=0.0, EW=E, alt=0.0, refDatum=W84)>",
            "<NMEA(GNRMC, time=10:36:07, status=A, lat=53.450657, NS=N, lon=-102.2404103333, EW=W, spd=0.046, cog=, date=2021-03-06, mv=, mvEW=, posMode=A, navStatus=V)>",
            "<NMEA(GPRTE, numMsg=2, msgNum=1, status=c, routeid=0, wpt_01=PBRCPK, wpt_02=PBRTO, wpt_03=PTELGR, wpt_04=PPLAND, wpt_05=PYAMBU, wpt_06=PPFAIR, wpt_07=PWARRN, wpt_08=PMORTL, wpt_09=PLISMR)>",
            "<NMEA(GNDTM, datum=W84, subDatum=, latOfset=0.0, NS=N, lonOfset=0.0, EW=E, alt=0.0, refDatum=W84)>",
            "<NMEA(GNRMC, time=10:36:07, status=A, lat=53.450657, NS=N, lon=-102.2404103333, EW=W, spd=0.046, cog=, date=2021-03-06, mv=, mvEW=, posMode=A, navStatus=V)>",
            "<NMEA(GPRTE, numMsg=2, msgNum=1, status=c, routeid=0, wpt_01=PBRCPK, wpt_02=PBRTO, wpt_03=PTELGR, wpt_04=PPLAND, wpt_05=PYAMBU, wpt_06=PPFAIR, wpt_07=PWARRN, wpt_08=PMORTL, wpt_09=PLISMR)>",
        )
        raw = None
        stream = DummySocket()
        nmr = NMEAReader(stream, bufsize=1024)
        buff = nmr._stream.buffer  # test buffer getter method
        i = 0
        for raw, parsed in nmr:
            if raw is not None:
                # print(parsed)
                self.assertEqual(str(parsed), EXPECTED_RESULTS[i])
                i += 1
                if i >= 12:
                    break
        self.assertEqual(i, 12)

    def testSocketSend(self):
        stream = DummySocket()
        nmr = NMEAReader(stream, bufsize=1024)
        msg = NMEAMessage("EI", "GNQ", POLL, msgId="RMC")
        res = nmr.datastream.write(msg.serialize())
        self.assertEqual(res, None)

    def testSocketIter(self):  # test for extended stream
        raw = None
        stream = DummySocket()
        nmr = NMEAReader(stream)
        i = 0
        for raw, parsed in nmr:
            if raw is None:
                raise EOFError
            i += 1
            if i >= 123:
                break
        self.assertEqual(i, 123)

    def testSocketError(self):  # test for simulated socket timeout
        raw = None
        stream = DummySocket(timeout=True)
        nmr = NMEAReader(stream)
        i = 0
        for raw, parsed in nmr:
            i += 1
            if i >= 12:
                break
        self.assertEqual(i, 0)


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



"""
NMEA Protocol Output (GET) payload definitions

THESE ARE THE PAYLOAD DEFINITIONS FOR _GET_ MESSAGES _FROM_ THE RECEIVER
(e.g. Periodic Navigation Data; Poll Responses; Info messages).

NB: Attribute names must be unique within each message id.
NB: Avoid reserved names 'msgID', 'talker', 'payload', 'checksum'.

NB: Repeating groups must be defined as a tuple thus
    'group': ('numr', {dict})
    where
    - 'numr' is either:
       a) an integer representing a fixed number of repeats e.g 32
       b) a string representing the name of a preceding attribute
          containing the number of repeats e.g. 'numCh'
       c) 'None' for an indeterminate repeating group
          (only one such group is permitted per message type)
    - {dict} is the nested dictionary containing the repeating
      attributes

Created on 4 Mar Sep 2021

While the NMEA 0183 © protocol is proprietary, the information here
has been collated from public domain sources.

:author: semuadmin (Steve Smith)
"""

# pylint: disable=too-many-lines, duplicate-code

from pynmeagps.nmeatypes_core import CH, DE, DT, DTL, HX, IN, LA, LAD, LN, LND, ST, TM

NMEA_PAYLOADS_GET = {
    # *********************************************
    # STANDARD MESSAGES
    # *********************************************
    "AAM": {
        "arrce": CH,
        "perp": CH,
        "crad": DE,
        "cUnit": CH,  # N nm
        "wpt": ST,
    },
    "ABK": {
        "mmsi": ST,
        "aischan": CH,
        "itumsg": ST,
        "msgSeq": IN,
        "ackType": IN,
    },
    "ABM": {
        "numSen": IN,
        "senNum": IN,
        "mmsi": ST,
        "aischan": CH,
        "ituMsg": ST,
        "data": ST,
        "fillbits": IN,
    },
    "ACA": {
        "seqNum": IN,
        "latne": LA,
        "latneNS": LAD,
        "lonne": LN,
        "lonneEW": LND,
        "latsw": LA,
        "latswNS": LAD,
        "lonswq": LN,
        "lonswEW": LND,
        "tranZoneSize": CH,
        "chanA": CH,
        "chanAbw": CH,
        "chanB": CH,
        "chanBbw": CH,
        "txrxMode": IN,
        "pwrLevelCtl": IN,
        "infoSource": CH,
        "inuse": IN,
        "inuseTime": TM,
    },
    "ACK": {
        "alarmid": ST,
    },
    "ACN": {
        "time": TM,
        "mfrcode": ST,
        "alertid": ST,
        "alertnum": IN,
        "alertcmd": ST,
        "status": CH,
    },
    "ACS": {
        "seqNum": IN,
        "mmsi": ST,
        "utc": TM,
        "day": IN,
        "month": IN,
        "year": IN,
    },
    "AIR": {
        "mmsi1": ST,
        "mmsi1msg1": ST,
        "mmsi1msg1sub": ST,
        "mmsi1msg2": ST,
        "mmsi1msg2sub": CH,
        "mmsi2": ST,
        "mmsi2msg1": ST,
        "mmsi2msg1sub": ST,
        "channel": ST,
        "mid11": ST,
        "mid12": ST,
        "mid21": ST,
    },
    "AKD": {
        "acktime": TM,
        "srcsystem": ST,
        "srcsubsysytem": ST,
        "srcinstance": ST,
        "alarmtype": ST,
        "acksystem": ST,
        "acksubsystem": ST,
        "ackinstance": ST,
    },
    "ALA": {
        "evttime": TM,
        "srcsystem": ST,
        "srcsubsysytem": ST,
        "srcinstance": ST,
        "alarmtype": ST,
        "alarmcond": ST,
        "alarmackstate": ST,
        "alarmtext": ST,
    },
    "ALC": {
        "numSen": IN,
        "senNum": IN,
        "seqmid": IN,
        "numAlerts": IN,
        "alertgroup": (  # repeating group * numAlerts
            "numAlerts",
            {
                "mfrcode": ST,
                "alertid": ST,
                "revisionctr": ST,
            },
        ),
    },
    "ALF": {
        "numSen": IN,
        "senNum": IN,
        "seqmid": IN,
        "timelastchange": TM,
        "alertcat": CH,  # A, B, C
        "alertpriority": CH,  # E, A W, C
        "alertstate": CH,  # A, S, N, O, U, V
        "mfrcode": ST,
        "alertid": ST,
        "alertinst": ST,
        "revisionctr": ST,
        "escalationctr": ST,
        "alerttxt": ST,  # max 16 chars
    },
    "ALR": {
        "timealarmchange": TM,
        "alarmnum": ST,
        "alarmcond": ST,
        "alarmstate": ST,
        "alarmtext": ST,
    },
    "APA": {
        "LCgwarn": CH,
        "LCcwarn": CH,
        "ctrkerr": DE,
        "dirs": CH,
        "ctrkUnit": CH,
        "aalmcirc": CH,
        "aalmperp": CH,
        "bearP2D": DE,
        "bearP2Du": CH,
        "wpt": ST,
    },
    "APB": {
        "LCgwarn": CH,
        "LCcwarn": CH,
        "ctrkerr": DE,
        "dirs": CH,
        "ctrkUnit": CH,
        "aalmcirc": CH,
        "aalmperp": CH,
        "bearO2D": DE,
        "bearO2Du": CH,
        "wpt": ST,
        "bearD": DE,
        "bearDu": CH,
        "bearS": DE,
        "bearSu": CH,
        "mode": CH,
    },
    "ARC": {
        "time": TM,
        "mfrcode": ST,
        "alertid": ST,
        "alertinst": IN,
        "alertcmd": ST,
    },
    "BBM": {
        "numSen": IN,
        "senNum": IN,
        "seqmid": IN,
        "aischan": CH,
        "itumsg": ST,
        "data": ST,
        "fillbits": IN,
    },
    "BEC": {
        "utc": TM,
        "wptlat": LA,
        "wptNS": LAD,
        "wptlon": LN,
        "wptEW": LND,
        "bearT": DE,
        "bearTu": CH,  # 'T'rue
        "bearM": DE,
        "bearMu": CH,  # 'M'agnetic
        "dist": DE,
        "distu": CH,  # 'N'autical Miles
        "wpt": ST,
    },
    "BOD": {
        "bearT": DE,
        "bearTu": CH,
        "bearM": DE,
        "bearMu": CH,
        "wptD": ST,
        "wptO": ST,
    },
    "BWC": {
        "fixutc": TM,
        "lat": LA,
        "NS": LAD,
        "lon": LN,
        "EW": LND,
        "bearT": DE,
        "bearTu": CH,  # 'T'
        "bearM": DE,
        "bearMu": CH,  # 'M'
        "dist": DE,
        "distUnit": CH,  # 'N'
        "wpt": ST,
        "mode": CH,
    },
    "BWR": {
        "fixutc": TM,
        "lat": LA,
        "NS": LAD,
        "lon": LN,
        "EW": LND,
        "bearT": DE,
        "bearTu": CH,  # 'T'
        "bearM": DE,
        "bearMu": CH,  # 'M'
        "dist": DE,
        "distUnit": CH,  # 'N'
        "wpt": ST,
        "mode": CH,
    },
    "BWW": {
        "bearT": DE,
        "bearTu": CH,  # 'T'
        "bearM": DE,
        "bearMu": CH,  # 'M'
        "wptto": ST,
        "wptfrom": ST,
    },
    "CUR": {
        "valid": CH,
        "datasetnum": IN,
        "layernum": IN,
        "depth": DE,
        "direction": DE,
        "dirTR": CH,  # 'T' or 'R'
        "speed": DE,
        "refdepth": DE,
        "heading": DE,
        "headingTM": CH,  # 'T' or 'M'
        "speedref": CH,  # 'B', 'W', 'P'
    },
    "DBT": {
        "depthf": DE,
        "depthfu": CH,  # 'f'eet
        "depthm": DE,
        "depthmu": CH,  # 'M'eter
        "depthfath": DE,
        "depthfathu": CH,  # 'F'athoms
    },
    "DDC": {
        "preset": CH,
        "brightness": IN,
        "palette": CH,
        "status": CH,
    },
    "DOR": {
        "msgtype": CH,
        "time": TM,
        "doormontype": ST,
        "div1ind": ST,
        "div2ind": ST,
        "doorcount": IN,
        "doorstatus": CH,
        "watertightdoorsetting": CH,
        "msgtext": ST,
    },
    "DPT": {
        "depth": DE,  # m
        "offset": DE,
        "rangescale": DE,
    },
    "DSC": {
        "format": ST,
        "address": ST,
        "category": ST,
        "distresstype": ST,
        "commtype": ST,
        "channel": ST,
        "timetelno": ST,
        "mmsi": ST,
        "distressnature": ST,
        "acknowledgement": CH,
        "expansionind": CH,
    },
    "DSE": {
        "numSen": IN,
        "senNum": IN,
        "queryflag": CH,
        "mmsi": ST,
        "datagroup": (
            "None",  # variable repeating group
            {
                "code": ST,
                "data": ST,
            },
        ),
    },
    "DTM": {
        "datum": ST,
        "subDatum": ST,
        "latOfset": DE,
        "NS": LAD,
        "lonOfset": DE,
        "EW": LND,
        "alt": DE,
        "refDatum": ST,
    },
    "EPV": {
        "status": CH,
        "equipmenttype": ST,
        "equipmentid": ST,
        "propertyid": ST,
        "value": ST,
    },
    "ETL": {
        "time": TM,
        "msgtype": CH,
        "engtelposind": ST,
        "engsubtelposind": ST,
        "operatinglocind": CH,
        "enginenum": CH,
    },
    "EVE": {
        "time": TM,
        "tagcode": ST,
        "eventtext": ST,
    },
    "FIR": {
        "msgtype": CH,
        "time": TM,
        "fdstype": ST,
        "div1ind": ST,
        "div2ind": ST,
        "fdsnum": ST,
        "condition": CH,
        "alarmackstate": CH,
        "text": ST,
    },
    # FMI from Feyman IME19EE Product Integration Guide v1.4.1
    "FMI": {
        "time": TM,
        "lat": DE,
        "lon": DE,
        "alt": DE,
        "roll": DE,
        "pitch": DE,
        "yaw": DE,
        "hrms": DE,
        "numSV": IN,
        "diffAge": IN,
        "navmod": IN,
        "status": HX,
    },
    "FSI": {
        "txfreq": ST,
        "rxfreq": ST,
        "opmode": ST,
        "power": IN,
        "status": CH,
    },
    "GBS": {
        "time": TM,
        "errLat": DE,
        "errLon": DE,
        "errAlt": DE,
        "svid": IN,
        "prob": DE,
        "bias": DE,
        "stddev": DE,
        "systemId": HX,  # NMEA >=4.10 only
        "signalId": HX,  # NMEA >=4.10 only
    },
    "GEN": {
        "index": HX,
        "time": TM,
        "group": ("None", {"data": HX}),  # variable repeating group
    },
    "GFA": {
        "time": TM,
        "hprotlvl": DE,
        "vprotlvl": DE,
        "semimajstddev": DE,
        "semiminstddev": DE,
        "semimajorientation": DE,
        "altstddev": DE,
        "accuracy": DE,
        "integrity": ST,  # V not in use, 'S'afe, 'C'aution, 'U'nsafe
    },
    "GGA": {
        "time": TM,
        "lat": LA,
        "NS": LAD,
        "lon": LN,
        "EW": LND,
        "quality": IN,
        "numSV": IN,
        "HDOP": DE,
        "alt": DE,  # altitude above sea level in m
        "altUnit": CH,
        "sep": DE,
        "sepUnit": CH,
        "diffAge": DE,
        "diffStation": IN,
    },
    "GLL": {
        "lat": LA,
        "NS": LAD,
        "lon": LN,
        "EW": LND,
        "time": TM,
        "status": ST,
        "posMode": ST,
    },
    "GMP": {
        "utctime": TM,
        "mapProjection": ST,
        "mapZone": ST,
        "gridN": DE,
        "gridE": DE,
        "modeInd": CH,
        "sip": IN,
        "HDOP": DE,
        "alt": DE,
        "geoidSep": DE,
        "corrAge": DE,
        "baseId": ST,
    },
    "GNS": {
        "time": TM,
        "lat": LA,
        "NS": LAD,
        "lon": LN,
        "EW": LND,
        "posMode": ST,
        "numSV": IN,
        "HDOP": DE,
        "alt": DE,
        "sep": DE,
        "diffAge": DE,
        "diffStation": IN,
        "navStatus": CH,  # NMEA >=4.10 only
    },
    "GRS": {
        "time": TM,
        "mode": IN,
        "groupSV": (
            12,
            {  # repeating group * 12
                "residual": DE,
            },
        ),
        "systemId": HX,  # NMEA >=4.10 only
        "signalId": HX,  # NMEA >=4.10 only
    },
    "GSA": {
        "opMode": CH,
        "navMode": IN,
        "groupSV": (
            12,
            {  # repeating group * 12
                "svid": IN,
            },
        ),
        "PDOP": DE,
        "HDOP": DE,
        "VDOP": DE,
        "systemId": HX,  # NMEA >=4.10 only
    },
    "GST": {
        "time": TM,
        "rangeRms": DE,
        "stdMajor": DE,
        "stdMinor": DE,
        "orient": DE,
        "stdLat": DE,
        "stdLong": DE,
        "stdAlt": DE,
    },
    "GSV": {
        "numMsg": IN,
        "msgNum": IN,
        "numSV": IN,
        "group_sv": (
            "None",
            {  # repeating group * 1..4
                "svid": IN,
                "elv": DE,  # elevation
                "az": IN,  # azimuth
                "cno": IN,  # signal strength
            },
        ),
        "signalID": HX,  # NMEA >=4.10 only
    },
    "HBT": {
        "repeatint": DE,
        "status": CH,
        "seqid": IN,
    },
    "HCR": {
        "headingt": DE,
        "mode": CH,
        "corrstate": CH,
        "corrvalue": DE,
    },
    "HDG": {
        "headingM": DE,
        "deviation": DE,
        "deviationEW": LND,
        "variation": DE,
        "variationEW": LND,
    },
    "HDM": {
        "headingM": DE,
        "deviation": DE,
        "deviationEW": LND,
        "variation": DE,
        "variationEW": LND,
    },
    "HDT": {
        "headingT": DE,
        "headingTu": CH,  # 'T'
    },
    "HMR": {
        "sensor1id": ST,
        "sensor2id": ST,
        "setdiff": DE,
        "actualdiff": DE,
        "warning": CH,  # A within set, V exceeds set
        "sensor1hdg": DE,
        "sensor1status": CH,
        "sensor1u": CH,  # 'T'rue, 'M'agnetic
        "sensor1dev": DE,
        "sensor1EW": LND,
        "sensor2hdg": DE,
        "sensor2status": CH,
        "sensor2u": CH,  # 'T'rue, 'M'agnetic
        "sensor2dev": DE,
        "sensor2EW": LND,
        "variation": DE,
        "variationEW": LND,
    },
    "HMS": {
        "sensor1id": ST,
        "sensor2id": ST,
        "setdiff": DE,
    },
    "HRM": {
        "heel": DE,
        "rollperiod": DE,
        "rollamplitudeport": DE,
        "rollamplitduestbd": DE,
        "status": CH,
        "rollpeakport": DE,
        "rollpeakstbd": DE,
        "peakholdresettime": TM,
        "peakholdresetdat": IN,
        "peakholdresetmonth": IN,
    },
    "HSC": {
        "cmdheadingT": DE,
        "cmdheadingTu": CH,  # 'T'
        "cmdheadingM": DE,
        "cmdheadingMu": CH,  # 'M'
        "status": CH,
    },
    "HSS": {
        "measpointid": ST,
        "value": DE,
        "status": CH,
    },
    "HTC": {
        "override": CH,  # A in use V not in use
        "cmdrudderang": DE,
        "cmdrudderdir": CH,  # L port R stbd
        "steermode": CH,
        "turnmode": CH,  # 'R'adius, 'T'urn rate, 'N'ot controlled
        "cmdrudderlimit": DE,
        "cmdoffheadinglimit": DE,
        "cmdturnradius": DE,
        "cmdturnrate": DE,
        "cmdheading": DE,
        "cmdofftracklimit": DE,
        "cmdtrack": DE,
        "headingref": CH,  # 'T'rue, 'M'agnetic
        "status": CH,
    },
    "HTD": {
        "override": CH,  # A in use V not in use
        "cmdrudderang": DE,
        "cmdrudderdir": CH,  # L port R stbd
        "steermode": CH,
        "turnmode": CH,  # 'R'adius, 'T'urn rate, 'N'ot controlled
        "cmdrudderlimit": DE,
        "cmdoffheadinglimit": DE,
        "cmdturnradius": DE,
        "cmdturnrate": DE,
        "cmdheading": DE,
        "cmdofftracklimit": DE,
        "cmdtrack": DE,
        "headingref": CH,  # 'T'rue, 'M'agnetic
        "rudderstatus": CH,
        "offheadingstatus": CH,
        "offtrackstatus": CH,
        "vesselheading": DE,
    },
    "LR1": {
        "seqNum": IN,
        "mmsiresponder": ST,
        "mmsirequester": ST,
        "shipname": ST,
        "callsign": ST,
        "imonum": ST,
    },
    "LR2": {
        "seqNum": IN,
        "mmsiresponder": ST,
        "date": DTL,
        "time": TM,
        "lat": LA,
        "NS": LAD,
        "lon": LN,
        "EW": LND,
        "cog": DE,  # course over ground
        "cogT": CH,  # 'T'
        "sog": DE,  # speed over ground
        "sogu": CH,  # 'N' knots
    },
    "LR3": {
        "seqNum": IN,
        "mmsiresponder": ST,
        "destination": ST,
        "etadate": DT,
        "etatime": TM,
        "draught": DE,
        "cargo": DE,
        "length": DE,
        "breadth": DE,
        "type": DE,
        "persons": DE,
    },
    "LRF": {
        "seqNum": IN,
        "mmsirequestor": ST,
        "namerequestor": ST,
        "function": ST,
        "replystatus": ST,
    },
    "LRI": {
        "seqNum": IN,
        "controlflag": CH,
        "mmsirequestor": ST,
        "mmsidestination": ST,
        "latne": LA,
        "NSne": LAD,
        "lonne": LN,
        "EWne": LND,
        "latsw": LA,
        "NSsw": LAD,
        "lonsw": LN,
        "EWsw": LND,
    },
    "LLQ": {
        "utctime": TM,
        "utcdate": DT,
        "easting": DE,
        "eunit": CH,  # 'M'
        "northing": DE,
        "nunit": CH,  # 'M'
        "gpsQual": IN,
        "sip": IN,
        "posQual": DE,
        "height": DE,
        "hunit": CH,  # 'M'
    },
    "MOB": {
        "emitterid": ST,
        "mobstatus": CH,
        "timeactivation": TM,
        "mobpossource": CH,
        "dayactivated": IN,
        "timeposition": TM,
        "lat": LA,
        "NS": LAD,
        "lon": LN,
        "EW": LND,
        "cog": DE,  # true
        "sog": DE,  # knots
        "mmsi": ST,
        "batterystatus": CH,
    },
    "MSK": {
        "freq": DE,
        "fmode": CH,  # 'A'uto, 'M'anual
        "beacbps": IN,
        "bpsmode": CH,  # 'A'uto, 'M'anual
        "MMSfreq": DE,
        "channel": CH,
        "status": CH,
    },
    "MSS": {
        "strength": IN,
        "snr": IN,
        "freq": DE,
        "beacbps": IN,
        "channel": CH,
    },
    "MTW": {
        "temp": DE,
        "tempu": CH,  # degrees C
    },
    "MWD": {
        "dirT": DE,
        "dirTu": CH,  # 'T'rue
        "dirM": DE,
        "dirMu": CH,  # 'M'agnetic
        "speedN": DE,
        "speedNu": CH,  # 'N' knots
        "speedM": DE,
        "speedMu": CH,  # 'M' m/s
    },
    "MWV": {
        "angle": DE,
        "reference": CH,  # 'R'elative, 'T'heoretical
        "speed": DE,
        "speedu": CH,  # K km/h, M m/s N knots
        "status": CH,
    },
    "NAK": {
        "talkerid": ST,
        "formatter": ST,
        "identifier": ST,
        "reason": ST,
        "text": ST,
    },
    "NRM": {
        "function": CH,
        "freqtable": CH,
        "txcovermask": HX,
        "msgtypemask": HX,
        "status": CH,
    },
    "NRX": {
        "numSen": IN,
        "senNum": IN,
        "seqid": IN,
        "msgcode": CH,
        "freqtable": CH,
        "time": TM,
        "day": IN,
        "month": IN,
        "year": IN,
        "totalchar": DE,
        "totalbad": DE,
        "status": CH,
        "body": ST,
    },
    "NSR": {
        "hdgintegrity": CH,
        "hdgplausibility": CH,
        "posintegrity": CH,
        "posplausibility": CH,
        "stwintegrity": CH,
        "stwplausibility": CH,
        "sogcogintegrity": CH,
        "sogcogplausibility": CH,
        "depthintegrity": CH,
        "depthplausibility": CH,
        "stwmode": CH,
        "timeintegrity": CH,
        "timeplausibility": CH,
    },
    "OSD": {
        "hdg": DE,
        "hdgstate": CH,
        "course": DE,
        "courseref": CH,
        "speed": DE,
        "speedref": CH,
        "set": DE,
        "drift": DE,
        "speedu": CH,  # K km/h N knots S mph
    },
    "POS": {
        "equipmentid": ST,
        "equipmentnum": IN,
        "posvalidity": CH,
        "posx": DE,
        "posy": DE,
        "posz": DE,
        "shipdimvalid": CH,
        "width": DE,
        "length": DE,
        "status": CH,
    },
    "PRC": {
        "leverdmdpos": DE,
        "leverdmdstatus": CH,
        "rpmdemand": DE,
        "rpmdemandstatus": CH,
        "pitchdmd": DE,
        "pitchmode": CH,
        "operatinglocind": CH,
        "engineshaftnum": IN,
    },
    "RLM": {
        "beacon": HX,
        "time": TM,
        "code": CH,
        "body": HX,
    },
    "RMA": {
        "status": CH,
        "lat": LA,
        "NS": LAD,
        "lon": LN,
        "EW": LND,
        "reserved1": ST,
        "reserved2": ST,
        "sog": DE,
        "cog": DE,
        "var": DE,
        "dirvar": CH,
    },
    "RMB": {
        "status": CH,
        "ctrkerr": DE,
        "dirs": CH,
        "wptO": CH,
        "wptD": CH,
        "lat": LA,  # of wptD
        "NS": LAD,
        "lon": LN,  # of wptD
        "EW": LND,
        "range": DE,
        "bearing": DE,
        "velclos": DE,
        "arrstatus": CH,
        "valstatus": CH,
    },
    "RMC": {
        "time": TM,
        "status": CH,
        "lat": LA,
        "NS": LAD,
        "lon": LN,
        "EW": LND,
        "spd": DE,  # speed in knots
        "cog": DE,  # course over ground
        "date": DT,
        "mv": DE,
        "mvEW": ST,
        "posMode": CH,
        "navStatus": CH,  # NMEA >=4.10 only
    },
    "ROR": {
        "stbdrudderorder": DE,
        "stbdrudderstatus": CH,
        "portrudderorder": DE,
        "portrudderstatus": CH,
        "operatinglocind": CH,
    },
    "ROT": {
        "rot": DE,  # -ve = turn to port
        "valid": CH,  # A valid, V invalid
    },
    "RRT": {
        "tfrtype": CH,
        "tfrroutename": ST,
        "tfrroutever": ST,
        "wptid": ST,
        "filestatus": CH,
        "appstatus": CH,
    },
    "RPM": {
        "source": CH,
        "engineshaftnum": IN,
        "rpm": DE,
        "pitch": DE,
        "status": CH,
    },
    "RSA": {
        "stbdangle": DE,
        "stbdstatus": CH,
        "portangle": DE,
        "portstatus": CH,
    },
    "RSD": {
        "origin1range": DE,
        "origin1bearing": DE,
        "vrm1": DE,
        "ebl1": DE,
        "origin2range": DE,
        "origin2bearing": DE,
        "vrm2": DE,
        "ebl2": DE,
        "cursorrange": DE,
        "rangebearing": DE,
        "rangescale": DE,
        "rangeu": CH,  # K kn N nautical mile S statute mile
        "displayrot": CH,
    },
    "RTE": {
        "numMsg": IN,
        "msgNum": IN,
        "status": CH,  # 'c'/'w'
        "routeid": ST,
        "group_wp": (
            "None",
            {  # repeating group
                "wpt": ST,
            },
        ),
    },
    "SFI": {
        "numSen": IN,
        "senNum": IN,
        "freqgroup": (
            6,  # repeating group * 6
            {
                "freq": IN,
                "mode": ST,
            },
        ),
    },
    "SM1": {
        "MSIstatus": CH,
        "msgnum": IN,
        "LESseqnum": IN,
        "LESid": IN,
        "oceanregioncode": IN,
        "prioritycode": IN,
        "servicecode": IN,
        "presentationcode": IN,
        "year": IN,
        "month": IN,
        "day": IN,
        "hour": IN,
        "minute": IN,
        "addresscode": IN,
    },
    "SM2": {
        "MSIstatus": CH,
        "msgnum": IN,
        "LESseqnum": IN,
        "LESid": IN,
        "oceanregioncode": IN,
        "prioritycode": IN,
        "servicecode": IN,
        "presentationcode": IN,
        "year": IN,
        "month": IN,
        "day": IN,
        "hour": IN,
        "minute": IN,
        "coastalwarnnavarea": IN,
        "coastalwarnarea": CH,
        "coastalwarnsubjectind": CH,
    },
    "SM3": {
        "MSIstatus": CH,
        "msgnum": IN,
        "LESseqnum": IN,
        "LESid": IN,
        "oceanregioncode": IN,
        "prioritycode": IN,
        "servicecode": IN,
        "presentationcode": IN,
        "year": IN,
        "month": IN,
        "day": IN,
        "hour": IN,
        "minute": IN,
        "circarealat": LA,
        "circareaNS": LAD,
        "circarealon": LN,
        "circareaEW": LND,
        "circarearadius": IN,
    },
    "SM4": {
        "MSIstatus": CH,
        "msgnum": IN,
        "LESseqnum": IN,
        "LESid": IN,
        "oceanregioncode": IN,
        "prioritycode": IN,
        "servicecode": IN,
        "presentationcode": IN,
        "year": IN,
        "month": IN,
        "day": IN,
        "hour": IN,
        "minute": IN,
        "redareaswlat": LA,
        "recareaswNS": LAD,
        "redareaswlon": LN,
        "recareaswEW": LND,
        "recareaextentN": IN,
        "recareaextentE": IN,
    },
    "SMB": {"numSen": IN, "senNum": IN, "seqid": IN, "msgnum": IN, "body": ST},
    "SPW": {
        "pwdprotectsentence": ST,
        "id": ST,
        "pwdlevel": IN,
        "password": ST,
    },
    "SSD": {
        "callsign": ST,
        "name": ST,
        "posAref": IN,  # bow
        "posBref": IN,  # stern
        "posCref": IN,  # port beam
        "posDref": IN,  # stbd beam
        "DTEind": ST,
        "sourceid": ST,
    },
    "STN": {
        "talkerId": ST,
    },
    "THS": {
        "headt": DE,
        "mi": CH,
    },
    "TLB": {
        "group": (
            "None",  # indeterminate repeating group
            {
                "targetnum": IN,
                "label": ST,
            },
        )
    },
    "TLL": {
        "targetnum": IN,
        "targetlat": LA,
        "targetlatNS": LAD,
        "targetlon": LN,
        "targetlonEW": LND,
        "targetname": ST,
        "time": TM,
        "status": CH,
        "reftarget": CH,
    },
    "TRC": {
        "thrusternum": IN,
        "dmdrpm": DE,
        "rpmmode": CH,
        "dmdpitch": DE,
        "pitchmode": CH,
        "dmdazimuth": DE,
        "operatinglocind": CH,
        "status": CH,
    },
    "TRD": {
        "thrusternum": IN,
        "rpmresponse": IN,
        "rpmmode": CH,
        "pitchresponse": IN,
        "pitchmode": CH,
        "azimuthresponse": IN,
    },
    "TRL": {
        "totlognum": IN,
        "lognum": IN,
        "seqid": IN,
        "switchoffdate": ST,
        "switchofftime": TM,
        "switchondate": ST,
        "switchontime": TM,
        "reasoncode": IN,
    },
    "TRF": {
        "time": TM,
        "date": DT,
        "lat": LA,
        "NS": LAD,
        "lon": LN,
        "EW": LND,
        "elangle": DE,
        "iter": DE,
        "Dopint": DE,
        "dist": DE,
        "svid": IN,
    },
    "TTD": {
        "numSen": IN,
        "senNum": IN,
        "seqid": IN,
        "data": ST,
        "fillbits": IN,
    },
    "TTM": {
        "targetnum": IN,
        "targetdist": DE,
        "targetbearing": DE,
        "targetbearingu": CH,  # 'T'rue, 'R'relative
        "targetspeed": DE,
        "targetcourse": DE,
        "targetcourseu": CH,  # 'T'rue, 'R'relative
        "closestapproach": DE,
        "timetoCPA": DE,
        "speedu": CH,
        "targetname": ST,
        "targetstatus": CH,
        "reftarget": ST,
        "time": TM,
        "acquisitiontype": CH,
    },
    "TUT": {
        "sourceid": ST,
        "numSen": HX,
        "senNum": HX,
        "seqid": IN,
        "translationcode": ST,
        "text": ST,
    },
    "TXT": {
        "numMsg": IN,
        "msgNum": IN,
        "msgType": IN,
        "text": ST,
    },
    "UID": {
        "uid1": ST,
        "uid2": ST,  # optional
    },
    "VBW": {
        "longwaterspd": DE,  # -ve = astern
        "transwaterspd": DE,
        "waterspdstatus": CH,
        "longgroundspd": DE,  # -ve = astern
        "transgroundspd": DE,  # -ve = port
        "groundspdstatus": CH,
        "sterntranswaterspd": DE,  # -ve = port
        "sternwaterspdstatus": CH,
        "sterntransgroundspd": DE,  # -ve = port
        "sterngroundspdstatus": CH,
    },
    "VDM": {
        "numSen": IN,
        "senNum": IN,
        "seqid": IN,
        "aischan": ST,
        "itumsg": ST,
        "fillbits": IN,
    },
    "VDO": {
        "numSen": IN,
        "senNum": IN,
        "seqid": IN,
        "aischan": ST,
        "itumsg": ST,
        "fillbits": IN,
    },
    "VDR": {
        "dirT": DE,
        "dirTu": CH,  # 'T'rue
        "dirM": DE,
        "dirMu": CH,  # 'M'agnetic
        "spd": DE,
        "spdu": CH,  # N knots
    },
    "VER": {
        "numSen": IN,
        "senNum": IN,
        "devicetype": ST,
        "vendorid": ST,
        "uid": ST,
        "mfrserial": ST,
        "model": ST,
        "software": ST,
        "hardware": ST,
        "seqid": IN,
    },
    "VHW": {
        "hdgT": DE,
        "hgdTu": CH,  # 'T'rue
        "hdgM": DE,
        "hgdMu": CH,  # 'M'agnetic
        "spdN": DE,
        "spdNu": CH,  # N knots
        "spdK": DE,
        "spdKu": CH,  # K km/h
    },
    "VLW": {
        "twd": DE,
        "twdUnit": CH,
        "wd": DE,
        "wdUnit": CH,
        "tgd": DE,  # NMEA >=4.00 only
        "tgdUnit": CH,  # NMEA >=4.00 only
        "gd": DE,  # NMEA >=4.00 only
        "gdUnit": CH,  # NMEA >=4.00 only
    },
    "VPW": {
        "spdN": DE,
        "spdNu": CH,  # N knots
        "spdK": DE,
        "spdKu": CH,  # K km/h
    },
    "VSD": {
        "type": ST,
        "draught": DE,
        "pob": IN,
        "destination": ST,
        "etatime": TM,
        "etaday": IN,
        "etamonth": IN,
        "regappflags1": IN,
        "regappflags2": IN,
    },
    "VTG": {
        "cogt": DE,  # course over ground (true)
        "cogtUnit": CH,
        "cogm": DE,  # course over ground (magnetic)
        "cogmUnit": CH,
        "sogn": DE,  # speed over ground knots
        "sognUnit": CH,
        "sogk": DE,  # speed over ground kmph
        "sogkUnit": CH,
        "posMode": CH,  # NMEA >=2.3 only
    },
    "WAT": {
        "msgtype": CH,
        "time": TM,
        "alarmtype": ST,
        "loc1ind": IN,
        "loc2ind": IN,
        "detectionpoint": IN,
        "alarmcondition": CH,
        "overridesetting": CH,
        "text": ST,
    },
    "WCV": {
        "vel": DE,
        "velu": CH,  # N knots
        "wpt": ST,
        "mode": CH,
    },
    "WNC": {
        "distN": DE,
        "distNu": CH,  # 'N' nm
        "distK": DE,
        "distKu": CH,  # 'K' km
        "wptto": ST,
        "wptfrom": ST,
    },
    "WPL": {
        "lat": LA,
        "NS": LAD,
        "lon": LN,
        "EW": LND,
        "wpt": ST,
    },
    "XDR": {
        "group": (
            "None",  # indeterminate repeating group
            {
                "tdrtype": CH,
                "data": DE,
                "dataunit": CH,
                "tdrid": ST,
            },
        )
    },
    "XTE": {
        "gwarn": CH,
        "LCcwarn": CH,
        "ctrkerr": DE,
        "dirs": CH,
        "disUnit": CH,  # N nm
        "mode": CH,
    },
    "XTR": {
        "ctkerr": DE,
        "steerdir": CH,
        "units": CH,  # N nm
    },
    "ZDA": {
        "time": TM,
        "day": IN,
        "month": IN,
        "year": IN,
        "ltzh": ST,
        "ltzn": ST,
    },
    "ZDL": {
        "time": TM,
        "dist": DE,
        "type": CH,
    },
    "ZFO": {
        "observationtime": TM,
        "elapsedtime": TM,
        "originwpt": ST,
    },
    "ZTG": {
        "observationtime": TM,
        "timetogo": TM,
        "destwpt": ST,
    },
    # *********************************************
    # Dummy message for error testing
    # *********************************************
    "FOO": {"spam": "Z2", "eggs": "Y1"},
}



"""
NMEA Protocol Proprietary Poll payload definitions

THESE ARE THE PAYLOAD DEFINITIONS FOR PROPRIETARY _POLL_ MESSAGES _TO_
THE RECEIVER (e.g. Message Poll requests).

NB: Attribute names must be unique within each message id.
NB: Avoid reserved names 'msgID', 'talker', 'payload', 'checksum'.

Created on 19 Aug 2024

While the NMEA 0183 © protocol is proprietary, the information here
has been collated from public domain sources.

:author: semuadmin (Steve Smith)
"""

from pynmeagps.nmeatypes_core import IN, QS, ST

NMEA_PAYLOADS_POLL_PROP = {
    # ***************************************************************
    # Quectel LC29H / LC79H Proprietary Poll message types
    # https://www.quectel.com/download/quectel_lc29hlc79h_series_gnss_protocol_specification_v1-5/
    # quectel_lc29hlc79h_series_gnss_protocol_specification_v1-5.pdf
    "AIR051": {},
    "AIR059": {},
    "AIR063": {"type": IN},
    "AIR067": {},
    "AIR071": {},
    "AIR073": {},
    "AIR075": {},
    "AIR081": {},
    "AIR087": {},
    "AIR101": {},
    "AIR105": {},
    "AIR401": {},
    "AIR411": {},
    "AIR421": {},
    "AIR433": {},
    "AIR435": {},
    "AIR865": {
        "portType": IN,  # 0 = UART
        "portIndex": IN,  # 0 = UART1
    },
    "AIR867": {
        "portType": IN,  # 0 = UART
        "portIndex": IN,  # 0 = UART1
    },
    "AIR6011": {
        "type": IN,
    },
    # ***************************************************************
    # Quectel LG290P Proprietary message types
    # https://quectel.com/content/uploads/2024/09/Quectel_LG290P03_GNSS_Protocol_Specification_V1.0.pdf
    #
    # status attribute must be set to 'R' in all POLL messages
    # ***************************************************************
    "QTMCFGAIC": {"status": QS},
    "QTMCFGCNST": {"status": QS},
    "QTMCFGDR": {"status": QS},
    "QTMCFGFIXRATE": {"status": QS},
    "QTMCFGGEOFENCE": {"status": QS, "geofenceindex": IN},
    "QTMCFGIMUINT": {"status": QS},
    "QTMCFGLA": {"status": QS},
    "QTMCFGLAM": {"status": QS},
    "QTMCFGMSGRATE_NOVER": {"status": QS, "msgname": ST},
    "QTMCFGMSGRATE": {"status": QS, "msgname": ST, "msgver": IN},
    "QTMCFGNAVMODE": {"status": QS},
    "QTMCFGNMEADP": {"status": QS},
    "QTMCFGNMEATID": {"status": QS},
    "QTMCFGODO": {"status": QS},
    "QTMCFGPPS": {"status": QS, "ppsindex": IN},
    "QTMCFGPROT": {"status": QS, "porttype": IN, "portid": IN},
    "QTMCFGRCVRMODE": {"status": QS},
    "QTMCFGRSID": {"status": QS},
    "QTMCFGRTCM": {"status": QS},
    "QTMCFGRTK": {"status": QS},
    "QTMCFGRTKSRCTYPE": {"status": QS},
    "QTMCFGSAT": {"status": QS, "systemid": IN, "signalid": ST},
    "QTMCFGSIGNAL": {"status": QS},
    "QTMCFGSTATICHOLD": {"status": QS},
    "QTMCFGSVIN": {"status": QS},
    "QTMCFGUART_CURR": {"status": QS},
    "QTMCFGUART": {"status": QS, "portid": IN},
    "QTMCFGVEHMOT": {"status": QS},
    "QTMGETUTC": {},
    "QTMQVER": {"msgver": IN},
    "QTMSN": {},
    "QTMUNIQID": {},
    "QTMVEHATT": {},
    "QTMVERNO": {},
}



"""
NMEA Protocol Poll payload definitions

THESE ARE THE PAYLOAD DEFINITIONS FOR _POLL_ MESSAGES _TO_ THE RECEIVER
(e.g. Message Poll requests).

NB: Attribute names must be unique within each message id.
NB: Avoid reserved names 'msgID', 'talker', 'payload', 'checksum'.

Created on 4 Mar Sep 2021

While the NMEA 0183 © protocol is proprietary, the information here
has been collated from public domain sources.

:author: semuadmin (Steve Smith)
"""

from pynmeagps.nmeatypes_core import ST

NMEA_PAYLOADS_POLL = {
    # *********************************************
    # STANDARD MESSAGES
    # *********************************************
    "GAQ": {  # Galileo
        "msgId": ST,
    },
    "GBQ": {  # BeiDou
        "msgId": ST,
    },
    "GLQ": {  # GLONASS
        "msgId": ST,
    },
    "GNQ": {  # Any GNSS
        "msgId": ST,
    },
    "GPQ": {  # GPS, SBAS
        "msgId": ST,
    },
    "GQQ": {  # QZSS
        "msgId": ST,
    },
    # *********************************************
    # PROPRIETARY MESSAGES
    # *********************************************
    # u-blox
    "UBX00": {
        "msgId": ST,  # '00'
    },
    "UBX03": {
        "msgId": ST,  # '03'
    },
    "UBX04": {
        "msgId": ST,  # '04'
    },
    "UBX05": {
        "msgId": ST,  # '05'
    },
    "UBX06": {
        "msgId": ST,  # '06'
    },
    # Garmin
    "GRMB": {},
    "GRMC": {},
    "GRMC1": {},
    "GRME": {},
    "GRMF": {},
    "GRMH": {},
    "GRMI": {},
    "GRMM": {},
    "GRMO": {},
    "GRMT": {},
    "GRMV": {},
    "GRMW": {},
    "GRMZ": {},
    # Trimble
    "ASHRALR": {
        "msgId": ST,  # 'ALR'
    },
    "ASHRARA": {
        "msgId": ST,  # 'ARA'
    },
    "ASHRARR": {
        "msgId": ST,  # 'ARR'
    },
    "ASHRATT": {
        "msgId": ST,  # 'ATT'
    },
    "ASHRBTS": {
        "msgId": ST,  # 'BTS'
    },
    "ASHRCAP": {
        "msgId": ST,  # 'CAP'
    },
    "ASHRCPA": {
        "msgId": ST,  # 'CPA'
    },
    "ASHRCPO": {
        "msgId": ST,  # 'CPO'
    },
    "ASHRDDM": {
        "msgId": ST,  # 'DDM'
    },
    "ASHRDDS": {
        "msgId": ST,  # 'DDS'
    },
    "ASHRHPR": {
        "msgId": ST,  # 'HPR'
    },
    "ASHRHR": {
        "msgId": ST,  # 'HR'
    },
    "ASHRLTN": {
        "msgId": ST,  # 'LTN'
    },
    "ASHRMDM": {
        "msgId": ST,  # 'MDM'
    },
    "ASHRPOS": {
        "msgId": ST,  # 'POS'
    },
    "ASHRPBN": {
        "msgId": ST,  # 'PBN'
    },
    "ASHRPTT": {
        "msgId": ST,  # 'PTT'
    },
    "ASHRPWR": {
        "msgId": ST,  # 'PWR'
    },
    "ASHRRCS": {
        "msgId": ST,  # 'RCS'
    },
    "ASHRSBD": {
        "msgId": ST,  # 'SBD'
    },
    "ASHRSGA": {
        "msgId": ST,  # 'SGA'
    },
    "ASHRSGL": {
        "msgId": ST,  # 'SGL'
    },
    "ASHRSGO": {
        "msgId": ST,  # 'SGO'
    },
    "ASHRSGP": {
        "msgId": ST,  # 'SGP'
    },
    "ASHRSIR": {
        "msgId": ST,  # 'SIR'
    },
    "ASHRSLB": {
        "msgId": ST,  # 'SLB'
    },
    "ASHRSQZ": {
        "msgId": ST,  # 'SQZ'
    },
    "ASHRSSB": {
        "msgId": ST,  # 'SSB'
    },
    "ASHRTEM": {
        "msgId": ST,  # 'TEM'
    },
    "ASHRTHS": {
        "msgId": ST,  # 'THS'
    },
    "ASHRTTT": {
        "msgId": ST,  # 'TTT'
    },
    "ASHRVCR": {
        "msgId": ST,  # 'VCR'
    },
    "ASHRVCT": {
        "msgId": ST,  # 'VCT'
    },
    "ASHRVEL": {
        "msgId": ST,  # 'VEL'
    },
    "FUGDP": {},
    "GPPADV110": {
        "msgId": ST,  # '110'
    },
    "GPPADV120": {
        "msgId": ST,  # '120'
    },
    "TNLAVR": {
        "msgId": ST,  # 'AVR'
    },
    "TNLBPQ": {
        "msgId": ST,  # 'BPQ'
    },
    "TNLDG": {
        "msgId": ST,  # 'DG'
    },
    "TNLEVT": {
        "msgId": ST,  # 'EVT'
    },
    "TNLGGK": {
        "msgId": ST,  # 'GGK'
    },
    "TNLGGKx": {
        "msgId": ST,  # 'GGKx'
    },
    "TNLPJK": {
        "msgId": ST,  # 'PJK'
    },
    "TNLPJT": {
        "msgId": ST,  # 'PJT'
    },
    "TNLREX": {
        "msgId": ST,  # 'REX'
    },
    "TNLVGK": {
        "msgId": ST,  # 'VGK'
    },
    "TNLVHD": {
        "msgId": ST,  # 'VHD'
    },
}



"""
NMEA Protocol Set payload definitions

THESE ARE THE PAYLOAD DEFINITIONS FOR _SET_ MESSAGES _TO_ THE RECEIVER
(e.g. Configuration commands).

NB: Attribute names must be unique within each message id.
NB: Avoid reserved names 'msgID', 'talker', 'payload', 'checksum'.

NB: Repeating groups must be defined as a tuple thus
    'group': ('numr', {dict})
    where
    - 'numr' is either:
       a) an integer representing a fixed number of repeats e.g 32
       b) a string representing the name of a preceding attribute
          containing the number of repeats e.g. 'numCh'
       c) 'None' for an indeterminate repeating group
          (only one such group is permitted per message type)
    - {dict} is the nested dictionary containing the repeating
      attributes

Created on 4 Mar Sep 2021

While the NMEA 0183 © protocol is proprietary, the information here
has been collated from public domain sources.

:author: semuadmin (Steve Smith)
"""

NMEA_PAYLOADS_SET = {
    # *********************************************
    # STANDARD MESSAGES
    # *********************************************
    # No standard SET messages that I'm aware of
}



"""
nmeatyps_decodes.py

NMEA Protocol attribute value decode constants.

From various public domain sources.

Created on 26 Aug 2023

:author: semuadmin (Steve Smith)
:copyright: semuadmin © 2021
:license: BSD 3-Clause
"""

FIXTYPE_GGA = QUALITY = {
    0: "NO FIX",
    1: "2D",
    2: "3D",
    3: "N/A",
    4: "RTK FIXED",
    5: "RTK FLOAT",
    6: "DR ONLY",
}
"""GGA quality/fix type values"""

GNSSLIST = {
    0: "GPS",
    1: "SBAS",
    2: "Galileo",
    3: "BeiDou",
    4: "IMES",
    5: "QZSS",
    6: "GLONASS",
    7: "NAVIC",
}
"""GNSS code"""

SYSTEMID = {
    "1": "GPS",
    "2": "GLONASS",
    "3": "Galileo",
    "4": "Beidou",
    "5": "QZSS",
    "6": "NavIC",
}
"""NMEA 4.1 System Identifier - key is systemId"""

SIGNALID = {
    ("1", "1"): "GPS L1C/A2",
    ("1", "6"): "GPS L2 CL",
    ("1", "5"): "GPS L2 CM",
    ("1", "7"): "GPS L5 I",
    ("1", "8"): "GPS L5 Q",
    ("3", "7"): "Galileo E1 B2/C2",
    ("3", "1"): "Galileo E5 aI/aQ",
    ("3", "2"): "Galileo E5 bI/bQ",
    ("4", "1"): "BeiDou B1I D12/D22",
    ("4", "B"): "BeiDou B2I D1/D2",
    ("4", "3"): "BeiDou B1 Cp/Cd",
    ("4", "5"): "BeiDou B2 ap/ad",
    ("5", "1"): "QZSS L1C/A2",
    ("5", "4"): "QZSS L1S",
    ("5", "5"): "QZSS L2 CM",
    ("5", "6"): "QZSS L2 CL",
    ("5", "7"): "QZSS L5 I",
    ("5", "8"): "QZSS L5 Q",
    ("2", "1"): "GLONASS L1 OF",
    ("2", "3"): "GLONASS L2 OF",
    ("6", "1"): "NavIC L5 A",
}
"""Signal Identifier - key is (systemId, signal Id), values are hex strings"""

FMI_STATUS = {
    0x00000001: ("Finit", "Filter uninitialized flag"),
    0x00000002: ("Ready", "Filter convergence completion flag"),
    0x00000004: ("Inaccurate", "Filter convergence is in process"),
    0x00000008: ("TiltReject", "Tilt angle over threshold"),
    0x00000010: ("GnssReject", "RTK data poor quality"),
    0x00000020: ("FReset", "Filter reset flag"),
    0x00000040: ("FixRlsStage1", "Installation angle estimation stage 1"),
    0x00000080: ("FixRlsStage2", "Installation angle estimation stage 2"),
    0x00000100: ("FixRlsStage3", "Installation angle estimation stage 3"),
    0x00000200: ("FixRlsStage4", "Installation angle estimation stage 4"),
    0x00000400: ("FixRlsOK", "Installation angle estimation complete"),
    0x00002000: ("Direction1", "Shake RTK pole to initialize direction 1"),
    0x00004000: ("Direction2", "Shake RTK pole to initialize direction 2"),
    0x00010000: ("GnssLost", "Invalid RTK data or RTK data missing"),
    0x00020000: ("FInitOk", "Filter initialization completed"),
    0x00040000: ("PPSReady", "PPS signal detected"),
    0x00080000: ("SyncReady", "Module time synchronized with PPS"),
    0x00100000: ("GnssConnect", "Serial connection with RTK established"),
}
"""GPFMI status values"""



"""
Stream method tests for pynmeagps

Created on 4 Mar 2021

*** NB: must be saved in UTF-8 format ***

:author: semuadmin (Steve Smith)
"""

import os
import sys
import unittest
from logging import ERROR

from pynmeagps import (
    NMEAReader,
    NMEAParseError,
    NMEATypeError,
    VALCKSUM,
    ERR_RAISE,
    ERR_IGNORE,
    ERR_LOG,
    GET,
    SET,
    TM,
    LA,
    LAD,
    LN,
    LND,
    ST,
    FMI_STATUS,
)

DIRNAME = os.path.dirname(__file__)


class StreamTest(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def tearDown(self):
        pass

    def catchio(self):
        """
        Capture stdout as string.
        """

        self._saved_stdout = sys.stdout
        self._strout = os.StringIO()
        sys.stdout = self._strout

    def restoreio(self) -> str:
        """
        Return captured output and restore stdout.
        """

        sys.stdout = self._saved_stdout
        return self._strout.getvalue().strip()

    def testNMEASTARTUP(self):  # stream of NMEA device during start up (blank data)
        EXPECTED_RESULTS = (
            "<NMEA(GNRMC, time=, status=V, lat=, NS=, lon=, EW=, spd=, cog=, date=, mv=, mvEW=, posMode=N, navStatus=V)>",
            "<NMEA(GNVTG, cogt=, cogtUnit=, cogm=, cogmUnit=, sogn=, sognUnit=, sogk=, sogkUnit=, posMode=N)>",
            "<NMEA(GNGGA, time=, lat=, NS=, lon=, EW=, quality=0, numSV=0, HDOP=99.99, alt=, altUnit=, sep=, sepUnit=, diffAge=, diffStation=)>",
            "<NMEA(GNGSA, opMode=A, navMode=1, svid_01=, svid_02=, svid_03=, svid_04=, svid_05=, svid_06=, svid_07=, svid_08=, svid_09=, svid_10=, svid_11=, svid_12=, PDOP=99.99, HDOP=99.99, VDOP=99.99, systemId=1)>",
            "<NMEA(GNGSA, opMode=A, navMode=1, svid_01=, svid_02=, svid_03=, svid_04=, svid_05=, svid_06=, svid_07=, svid_08=, svid_09=, svid_10=, svid_11=, svid_12=, PDOP=99.99, HDOP=99.99, VDOP=99.99, systemId=2)>",
            "<NMEA(GNGSA, opMode=A, navMode=1, svid_01=, svid_02=, svid_03=, svid_04=, svid_05=, svid_06=, svid_07=, svid_08=, svid_09=, svid_10=, svid_11=, svid_12=, PDOP=99.99, HDOP=99.99, VDOP=99.99, systemId=3)>",
            "<NMEA(GNGSA, opMode=A, navMode=1, svid_01=, svid_02=, svid_03=, svid_04=, svid_05=, svid_06=, svid_07=, svid_08=, svid_09=, svid_10=, svid_11=, svid_12=, PDOP=99.99, HDOP=99.99, VDOP=99.99, systemId=4)>",
            "<NMEA(GPGSV, numMsg=1, msgNum=1, numSV=0, signalID=1)>",
            "<NMEA(GLGSV, numMsg=1, msgNum=1, numSV=0, signalID=1)>",
            "<NMEA(GAGSV, numMsg=1, msgNum=1, numSV=0, signalID=7)>",
            "<NMEA(GBGSV, numMsg=1, msgNum=1, numSV=0, signalID=1)>",
            "<NMEA(GNGLL, lat=, NS=, lon=, EW=, time=, status=V, posMode=N)>",
        )

        i = 0
        raw = 0
        with open(os.path.join(DIRNAME, "pygpsdata-nmeastartup.log"), "rb") as stream:
            nmr = NMEAReader(stream, nmeaonly=False, validate=1)
            while raw is not None:
                (raw, parsed) = nmr.read()
                if raw is not None:
                    self.assertEqual(str(parsed), EXPECTED_RESULTS[i])
                    i += 1
        self.assertEqual(
            i, 12
        )  # if this fails, may be because log file terminators = LF rather than CRLF

    def testNMEA4(
        self,
    ):  # stream of NMEA v4.10 device (u-blox M9N) (NB everything after PUBX is synthetic)
        EXPECTED_RESULTS = (
            "<NMEA(GNDTM, datum=W84, subDatum=, latOfset=0.0, NS=N, lonOfset=0.0, EW=E, alt=0.0, refDatum=W84)>",
            "<NMEA(GNRMC, time=10:36:07, status=A, lat=53.450657, NS=N, lon=-102.2404103333, EW=W, spd=0.046, cog=, date=2021-03-06, mv=, mvEW=, posMode=A, navStatus=V)>",
            "<NMEA(GPRTE, numMsg=2, msgNum=1, status=c, routeid=0, wpt_01=PBRCPK, wpt_02=PBRTO, wpt_03=PTELGR, wpt_04=PPLAND, wpt_05=PYAMBU, wpt_06=PPFAIR, wpt_07=PWARRN, wpt_08=PMORTL, wpt_09=PLISMR)>",
            "<NMEA(GNRLM, beacon=00000078A9FBAD5, time=08:35:59, code=3, body=C45B)>",
            "<NMEA(GNVTG, cogt=, cogtUnit=T, cogm=, cogmUnit=M, sogn=0.046, sognUnit=N, sogk=0.085, sogkUnit=K, posMode=A)>",
            "<NMEA(GNGNS, time=10:36:07, lat=53.450657, NS=N, lon=-2.2404103333, EW=W, posMode=AANN, numSV=6, HDOP=5.88, alt=56.0, sep=48.5, diffAge=, diffStation=, navStatus=V)>",
            "<NMEA(GNGGA, time=10:36:07, lat=53.450657, NS=N, lon=-2.2404103333, EW=W, quality=1, numSV=6, HDOP=5.88, alt=56.0, altUnit=M, sep=48.5, sepUnit=M, diffAge=, diffStation=)>",
            "<NMEA(GNGSA, opMode=A, navMode=3, svid_01=23, svid_02=24, svid_03=20, svid_04=12, svid_05=, svid_06=, svid_07=, svid_08=, svid_09=, svid_10=, svid_11=, svid_12=, PDOP=9.62, HDOP=5.88, VDOP=7.62, systemId=1)>",
            "<NMEA(GNGSA, opMode=A, navMode=3, svid_01=66, svid_02=76, svid_03=, svid_04=, svid_05=, svid_06=, svid_07=, svid_08=, svid_09=, svid_10=, svid_11=, svid_12=, PDOP=9.62, HDOP=5.88, VDOP=7.62, systemId=2)>",
            "<NMEA(GNGSA, opMode=A, navMode=3, svid_01=, svid_02=, svid_03=, svid_04=, svid_05=, svid_06=, svid_07=, svid_08=, svid_09=, svid_10=, svid_11=, svid_12=, PDOP=9.62, HDOP=5.88, VDOP=7.62, systemId=3)>",
            "<NMEA(GNGSA, opMode=A, navMode=3, svid_01=, svid_02=, svid_03=, svid_04=, svid_05=, svid_06=, svid_07=, svid_08=, svid_09=, svid_10=, svid_11=, svid_12=, PDOP=9.62, HDOP=5.88, VDOP=7.62, systemId=4)>",
            "<NMEA(GPGSV, numMsg=3, msgNum=1, numSV=11, svid_01=1, elv_01=6.0, az_01=14, cno_01=8, svid_02=12, elv_02=43.0, az_02=207, cno_02=28, svid_03=14, elv_03=6.0, az_03=49, cno_03=, svid_04=15, elv_04=44.0, az_04=171, cno_04=23, signalID=1)>",
            "<NMEA(GPGSV, numMsg=3, msgNum=2, numSV=11, svid_01=17, elv_01=32.0, az_01=64, cno_01=16, svid_02=19, elv_02=33.0, az_02=94, cno_02=, svid_03=20, elv_03=20.0, az_03=251, cno_03=31, svid_04=21, elv_04=4.0, az_04=354, cno_04=, signalID=1)>",
            "<NMEA(GPGSV, numMsg=3, msgNum=3, numSV=11, svid_01=23, elv_01=27.0, az_01=251, cno_01=31, svid_02=24, elv_02=89.0, az_02=268, cno_02=26, svid_03=25, elv_03=5.0, az_03=223, cno_03=, signalID=1)>",
            "<NMEA(GLGSV, numMsg=3, msgNum=1, numSV=10, svid_01=65, elv_01=7.0, az_01=176, cno_01=, svid_02=66, elv_02=57.0, az_02=223, cno_02=35, svid_03=67, elv_03=42.0, az_03=315, cno_03=23, svid_04=68, elv_04=0.0, az_04=341, cno_04=29, signalID=1)>",
            "<NMEA(GLGSV, numMsg=3, msgNum=2, numSV=10, svid_01=75, elv_01=37.0, az_01=57, cno_01=, svid_02=76, elv_02=78.0, az_02=303, cno_02=18, svid_03=77, elv_03=27.0, az_03=253, cno_03=21, svid_04=84, elv_04=19.0, az_04=18, cno_04=, signalID=B)>",
            "<NMEA(GLGSV, numMsg=3, msgNum=3, numSV=10, svid_01=85, elv_01=22.0, az_01=78, cno_01=, svid_02=86, elv_02=1.0, az_02=121, cno_02=, signalID=1)>",
            "<NMEA(GAGSV, numMsg=1, msgNum=1, numSV=0, signalID=7)>",
            "<NMEA(GBGSV, numMsg=1, msgNum=1, numSV=2, svid_01=21, elv_01=, az_01=, cno_01=15, svid_02=25, elv_02=, az_02=, cno_02=28, signalID=1)>",
            "<NMEA(GNGLL, lat=53.450657, NS=N, lon=-2.2404103333, EW=W, time=10:36:07, status=A, posMode=A)>",
            "<NMEA(GNGRS, time=10:36:07, mode=1, residual_01=-2.1, residual_02=0.2, residual_03=2.7, residual_04=-0.4, residual_05=, residual_06=, residual_07=, residual_08=, residual_09=, residual_10=, residual_11=, residual_12=, systemId=1, signalId=1)>",
            "<NMEA(GNGRS, time=10:36:07, mode=1, residual_01=0.6, residual_02=5.1, residual_03=, residual_04=, residual_05=, residual_06=, residual_07=, residual_08=, residual_09=, residual_10=, residual_11=, residual_12=, systemId=2, signalId=1)>",
            "<NMEA(GNGRS, time=10:36:07, mode=1, residual_01=, residual_02=, residual_03=, residual_04=, residual_05=, residual_06=, residual_07=, residual_08=, residual_09=, residual_10=, residual_11=, residual_12=, systemId=3, signalId=7)>",
            "<NMEA(GNGRS, time=10:36:07, mode=1, residual_01=, residual_02=, residual_03=, residual_04=, residual_05=, residual_06=, residual_07=, residual_08=, residual_09=, residual_10=, residual_11=, residual_12=, systemId=4, signalId=1)>",
            "<NMEA(GNGST, time=10:36:07, rangeRms=38.0, stdMajor=60.0, stdMinor=38.0, orient=89.0, stdLat=15.0, stdLong=24.0, stdAlt=31.0)>",
            "<NMEA(GNZDA, time=10:36:07, day=6, month=3, year=2021, ltzh=00, ltzn=00)>",
            "<NMEA(GNGBS, time=10:36:07, errLat=15.1, errLon=24.2, errAlt=31.0, svid=, prob=, bias=, stddev=, systemId=, signalId=)>",
            "<NMEA(GNVLW, twd=, twdUnit=N, wd=, wdUnit=N, tgd=0.0, tgdUnit=N, gd=0.0, gdUnit=N)>",
            "<NMEA(PUBX00, msgId=00, time=10:36:07, lat=53.450657, NS=N, lon=-2.2404103333, EW=W, altRef=104.461, navStat=G3, hAcc=29.0, vAcc=31.0, SOG=0.085, COG=39.63, vVel=-0.007, diffAge=, HDOP=5.88, VDOP=7.62, TDOP=8.09, numSVs=6, reserved=0, DR=0)>",
            "<NMEA(PUBX03, msgId=03, numSv=23, svid_01=1, status_01=-, azi_01=14.0, ele_01=6.0, cno_01=8, lck_01=0, svid_02=12, status_02=U, azi_02=207.0, ele_02=43.0, cno_02=28, lck_02=9, svid_03=14, status_03=-, azi_03=49.0, ele_03=6.0, cno_03=, lck_03=0, svid_04=15, status_04=-, azi_04=171.0, ele_04=44.0, cno_04=23, lck_04=0, svid_05=17, status_05=-, azi_05=64.0, ele_05=32.0, cno_05=16, lck_05=0, svid_06=19, status_06=-, azi_06=94.0, ele_06=33.0, cno_06=, lck_06=0, svid_07=20, status_07=U, azi_07=251.0, ele_07=20.0, cno_07=31, lck_07=38, svid_08=21, status_08=-, azi_08=354.0, ele_08=4.0, cno_08=, lck_08=0, svid_09=23, status_09=U, azi_09=251.0, ele_09=27.0, cno_09=31, lck_09=64, svid_10=24, status_10=U, azi_10=268.0, ele_10=89.0, cno_10=26, lck_10=0, svid_11=25, status_11=-, azi_11=223.0, ele_11=5.0, cno_11=, lck_11=0, svid_12=48, status_12=-, azi_12=, ele_12=, cno_12=15, lck_12=0, svid_13=52, status_13=-, azi_13=, ele_13=, cno_13=28, lck_13=13, svid_14=65, status_14=-, azi_14=176.0, ele_14=7.0, cno_14=, lck_14=0, svid_15=66, status_15=U, azi_15=223.0, ele_15=57.0, cno_15=35, lck_15=64, svid_16=67, status_16=-, azi_16=315.0, ele_16=42.0, cno_16=23, lck_16=0, svid_17=68, status_17=-, azi_17=341.0, ele_17=0.0, cno_17=29, lck_17=0, svid_18=75, status_18=-, azi_18=57.0, ele_18=37.0, cno_18=, lck_18=0, svid_19=76, status_19=U, azi_19=303.0, ele_19=78.0, cno_19=18, lck_19=0, svid_20=77, status_20=-, azi_20=253.0, ele_20=27.0, cno_20=21, lck_20=0, svid_21=84, status_21=-, azi_21=18.0, ele_21=19.0, cno_21=, lck_21=0, svid_22=85, status_22=-, azi_22=78.0, ele_22=22.0, cno_22=, lck_22=0, svid_23=86, status_23=-, azi_23=121.0, ele_23=1.0, cno_23=, lck_23=0)>",
            "<NMEA(PUBX04, msgId=04, time=10:36:07, date=2021-03-06, utcTow=556567.00, utcWk=2147, leapSec=18, clkBias=-384839.0, clkDrift=-53.623, tpGran=16)>",
            "<NMEA(GPWPL, lat=49.286, NS=N, lon=-123.1773333333, EW=W, wpt=003)>",
            "<NMEA(GPRMA, status=A, lat=53.450657, NS=N, lon=-112.2404103333, EW=W, reserved1=, reserved2=, sog=23.1, cog=23.0, var=14.8, dirvar=W)>",
            "<NMEA(GPRMB, status=A, ctrkerr=0.66, dirs=L, wptO=003, wptD=004, lat=49.2873333333, NS=N, lon=-123.1595, EW=W, range=1.3, bearing=52.5, velclos=0.5, arrstatus=V)>",
            "<NMEA(PGRME, HPE=15.0, HPEUnit=M, VPE=45.0, VPEUnit=M, EPE=25.0, EPEUnit=M)>",
            "<NMEA(PGRMM, dtm=NAD27 Canada)>",
            "<NMEA(PGRMZ, alt=246.0, altUnit=f, fix=3)>",
            "<NMEA(GPXTE, gwarn=A, LCcwarn=A, ctrkerr=4.07, dirs=L, disUnit=N)>",
            "<NMEA(GPVBW, longwaterspd=12.3, transwaterspd=0.07, waterspdstatus=A, longgroundspd=11.78, transgroundspd=0.12, groundspdstatus=A)>",
            "<NMEA(GPSTN, talkerId=34)>",
            "<NMEA(GPBWC, fixutc=22:05:16, lat=51.5003333333, NS=N, lon=-0.7723333333, EW=W, bearT=213.8, bearTu=T, bearM=218.0, bearMu=M, dist=4.6, distUnit=N, wpt=EGLM)>",
            "<NMEA(GPBOD, bearT=97.0, bearTu=T, bearM=103.2, bearMu=M, wptD=POINTB, wptO=POINTA)>",
            "<NMEA(GPBOD, bearT=99.3, bearTu=T, bearM=105.6, bearMu=M, wptD=POINTB, wptO=)>",
            "<NMEA(GPAAM, arrce=A, perp=A, crad=0.1, cUnit=N, wpt=WPTNME)>",
            "<NMEA(GPAPB, LCgwarn=A, LCcwarn=A, ctrkerr=0.1, dirs=R, ctrkUnit=N, aalmcirc=V, aalmperp=V, bearO2D=11.0, bearO2Du=M, wpt=DEST, bearD=11.0, bearDu=M, bearS=11.0, bearSu=M)>",
            "<NMEA(GPMSK, freq=318.0, fmode=A, beacbps=100, bpsmode=M, MMSfreq=2.0)>",
            "<NMEA(GPMSS, strength=55, snr=27, freq=318.0, beacbps=100, channel=)>",
            "<NMEA(GBGSV, numMsg=2, msgNum=2, numSV=6, svid_01=14, elv_01=55.0, az_01=175, cno_01=46, svid_02=40, elv_02=29.0, az_02=43, cno_02=18, signalID=B)>",
            "<NMEA(INGGA, time=10:36:07, lat=53.450657, NS=N, lon=-2.2404103333, EW=W, quality=1, numSV=6, HDOP=5.88, alt=56.0, altUnit=M, sep=48.5, sepUnit=M, diffAge=, diffStation=)>",
            "<NMEA(ANNAK, talkerid=UP, formatter=AID, identifier=999999990, reason=11, text=)>",
            "<NMEA(GPGBS, time=23:55:03, errLat=1.6, errLon=1.4, errAlt=3.2, svid=, prob=, bias=, stddev=, systemId=, signalId=)>",
            "<NMEA(GPGBS, time=23:54:58, errLat=1.4, errLon=1.3, errAlt=3.1, svid=3, prob=, bias=-21.4, stddev=3.8, systemId=1, signalId=0)>",
        )

        i = 0
        raw = 0
        with open(os.path.join(DIRNAME, "pygpsdata-nmea4.log"), "rb") as stream:
            nmr = NMEAReader(stream, nmeaonly=False, validate=3)
            while raw is not None:
                (raw, parsed) = nmr.read()
                if raw is not None:
                    # print(f'"{parsed}",')
                    self.assertEqual(str(parsed), EXPECTED_RESULTS[i])
                    i += 1
        self.assertEqual(i, len(EXPECTED_RESULTS))

    def testNMEA2(self):  # stream of NMEA v2.30 device (u-blox M6N)
        EXPECTED_RESULTS = (
            "<NMEA(GPTXT, numMsg=1, msgNum=1, msgType=2, text=u-blox ag - www.u-blox.com)>",
            "<NMEA(GPTXT, numMsg=1, msgNum=1, msgType=2, text=HW  UBX-G70xx   00070000 )>",
            "<NMEA(GPTXT, numMsg=1, msgNum=1, msgType=2, text=ROM CORE 1.00 (59842) Jun 27 2012 17:43:52)>",
            "<NMEA(GPTXT, numMsg=1, msgNum=1, msgType=2, text=PROTVER 14.00)>",
            "<NMEA(GPTXT, numMsg=1, msgNum=1, msgType=2, text=ANTSUPERV=AC SD PDoS SR)>",
            "<NMEA(GPTXT, numMsg=1, msgNum=1, msgType=2, text=ANTSTATUS=OK)>",
            "<NMEA(GPTXT, numMsg=1, msgNum=1, msgType=2, text=LLC FFFFFFFF-FFFFFFFD-FFFFFFFF-FFFFFFFF-FFFFFFF9)>",
            "<NMEA(GPRMC, time=10:29:29, status=A, lat=53.4506706667, NS=N, lon=-2.24026, EW=W, spd=0.273, cog=, date=2021-03-07, mv=, mvEW=, posMode=A)>",
            "<NMEA(GPVTG, cogt=, cogtUnit=T, cogm=, cogmUnit=M, sogn=0.273, sognUnit=N, sogk=0.506, sogkUnit=K, posMode=A)>",
            "<NMEA(GPGGA, time=10:29:29, lat=53.4506706667, NS=N, lon=-2.24026, EW=W, quality=1, numSV=8, HDOP=1.16, alt=36.3, altUnit=M, sep=48.5, sepUnit=M, diffAge=, diffStation=)>",
            "<NMEA(GPGSA, opMode=A, navMode=3, svid_01=17, svid_02=15, svid_03=10, svid_04=24, svid_05=20, svid_06=12, svid_07=19, svid_08=23, svid_09=, svid_10=, svid_11=, svid_12=, PDOP=2.36, HDOP=1.16, VDOP=2.05)>",
            "<NMEA(GPGSV, numMsg=4, msgNum=1, numSV=15, svid_01=1, elv_01=6.0, az_01=15, cno_01=, svid_02=10, elv_02=30.0, az_02=290, cno_02=27, svid_03=12, elv_03=42.0, az_03=207, cno_03=26, svid_04=13, elv_04=19.0, az_04=141, cno_04=23)>",
            "<NMEA(GPGSV, numMsg=4, msgNum=2, numSV=15, svid_01=14, elv_01=7.0, az_01=49, cno_01=21, svid_02=15, elv_02=45.0, az_02=171, cno_02=27, svid_03=17, elv_03=32.0, az_03=65, cno_03=22, svid_04=19, elv_04=33.0, az_04=95, cno_04=25)>",
            "<NMEA(GPGSV, numMsg=4, msgNum=3, numSV=15, svid_01=20, elv_01=21.0, az_01=251, cno_01=31, svid_02=21, elv_02=4.0, az_02=355, cno_02=, svid_03=23, elv_03=28.0, az_03=252, cno_03=33, svid_04=24, elv_04=88.0, az_04=273, cno_04=36)>",
            "<NMEA(GPGSV, numMsg=4, msgNum=4, numSV=15, svid_01=25, elv_01=5.0, az_01=223, cno_01=, svid_02=28, elv_02=14.0, az_02=49, cno_02=26, svid_03=32, elv_03=10.0, az_03=313, cno_03=16)>",
            "<NMEA(GPGLL, lat=53.4506706667, NS=N, lon=-2.24026, EW=W, time=10:29:29, status=A, posMode=A)>",
            "<NMEA(GPRMC, time=10:29:30, status=A, lat=53.4506721667, NS=N, lon=-2.2402583333, EW=W, spd=0.099, cog=, date=2021-03-07, mv=, mvEW=, posMode=A)>",
        )

        i = 0
        raw = 0
        with open(os.path.join(DIRNAME, "pygpsdata-nmea2.log"), "rb") as stream:
            nmr = NMEAReader(stream, nmeaonly=False)
            while raw is not None:
                (raw, parsed) = nmr.read()
                if raw is not None:
                    self.assertEqual(str(parsed), EXPECTED_RESULTS[i])
                    i += 1
        self.assertEqual(i, 17)

    def testMIXED(
        self,
    ):  # stream of mixed NMEA & UBX data with nmea_only set to FALSE - should be ignored
        EXPECTED_RESULTS = (
            "<NMEA(GNGGA, time=10:41:13, lat=53.4505928333, NS=N, lon=-2.2403723333, EW=W, quality=1, numSV=5, HDOP=8.68, alt=65.4, altUnit=M, sep=48.5, sepUnit=M, diffAge=, diffStation=)>",
            "<NMEA(GNGSA, opMode=A, navMode=3, svid_01=20, svid_02=10, svid_03=23, svid_04=, svid_05=, svid_06=, svid_07=, svid_08=, svid_09=, svid_10=, svid_11=, svid_12=, PDOP=12.55, HDOP=8.68, VDOP=9.07, systemId=1)>",
            "<NMEA(GNGSA, opMode=A, navMode=3, svid_01=78, svid_02=68, svid_03=, svid_04=, svid_05=, svid_06=, svid_07=, svid_08=, svid_09=, svid_10=, svid_11=, svid_12=, PDOP=12.55, HDOP=8.68, VDOP=9.07, systemId=2)>",
            "<NMEA(GNGSA, opMode=A, navMode=3, svid_01=, svid_02=, svid_03=, svid_04=, svid_05=, svid_06=, svid_07=, svid_08=, svid_09=, svid_10=, svid_11=, svid_12=, PDOP=12.55, HDOP=8.68, VDOP=9.07, systemId=3)>",
            "<NMEA(GNGSA, opMode=A, navMode=3, svid_01=, svid_02=, svid_03=, svid_04=, svid_05=, svid_06=, svid_07=, svid_08=, svid_09=, svid_10=, svid_11=, svid_12=, PDOP=12.55, HDOP=8.68, VDOP=9.07, systemId=4)>",
            "<NMEA(GPGSV, numMsg=1, msgNum=1, numSV=4, svid_01=10, elv_01=29.0, az_01=284, cno_01=30, svid_02=20, elv_02=17.0, az_02=247, cno_02=36, svid_03=23, elv_03=24.0, az_03=247, cno_03=38, svid_04=28, elv_04=10.0, az_04=49, cno_04=7, signalID=1)>",
            "<NMEA(GLGSV, numMsg=2, msgNum=1, numSV=7, svid_01=66, elv_01=10.0, az_01=179, cno_01=, svid_02=67, elv_02=52.0, az_02=218, cno_02=22, svid_03=68, elv_03=46.0, az_03=309, cno_03=30, svid_04=69, elv_04=5.0, az_04=340, cno_04=17, signalID=1)>",
            "<NMEA(GLGSV, numMsg=2, msgNum=2, numSV=7, svid_01=78, elv_01=30.0, az_01=255, cno_01=33, svid_02=85, elv_02=19.0, az_02=18, cno_02=, svid_03=86, elv_03=25.0, az_03=71, cno_03=, signalID=1)>",
            "<NMEA(GAGSV, numMsg=1, msgNum=1, numSV=0, signalID=7)>",
            "<NMEA(GBGSV, numMsg=1, msgNum=1, numSV=0, signalID=1)>",
            "<NMEA(GNGGA, time=10:41:14, lat=53.4505926667, NS=N, lon=-2.240361, EW=W, quality=1, numSV=5, HDOP=8.68, alt=65.2, altUnit=M, sep=48.5, sepUnit=M, diffAge=, diffStation=)>",
            "<NMEA(GNGSA, opMode=A, navMode=3, svid_01=20, svid_02=10, svid_03=23, svid_04=, svid_05=, svid_06=, svid_07=, svid_08=, svid_09=, svid_10=, svid_11=, svid_12=, PDOP=12.55, HDOP=8.68, VDOP=9.06, systemId=1)>",
            "<NMEA(GNGSA, opMode=A, navMode=3, svid_01=78, svid_02=68, svid_03=, svid_04=, svid_05=, svid_06=, svid_07=, svid_08=, svid_09=, svid_10=, svid_11=, svid_12=, PDOP=12.55, HDOP=8.68, VDOP=9.06, systemId=2)>",
            "<NMEA(GNGSA, opMode=A, navMode=3, svid_01=, svid_02=, svid_03=, svid_04=, svid_05=, svid_06=, svid_07=, svid_08=, svid_09=, svid_10=, svid_11=, svid_12=, PDOP=12.55, HDOP=8.68, VDOP=9.06, systemId=3)>",
            "<NMEA(GNGSA, opMode=A, navMode=3, svid_01=, svid_02=, svid_03=, svid_04=, svid_05=, svid_06=, svid_07=, svid_08=, svid_09=, svid_10=, svid_11=, svid_12=, PDOP=12.55, HDOP=8.68, VDOP=9.06, systemId=4)>",
        )

        i = 0
        raw = 0
        with open(os.path.join(DIRNAME, "pygpsdata-mixed.log"), "rb") as stream:
            nmr = NMEAReader(stream, nmeaonly=False)
            while raw is not None:
                (raw, parsed) = nmr.read()
                if raw is not None:
                    self.assertEqual(str(parsed), EXPECTED_RESULTS[i])
                    i += 1
        self.assertEqual(i, 15)

    def testMIXED2(
        self,
    ):  # stream of mixed NMEA & UBX data with nmea_only set to TRUE - should be rejected
        EXPECTED_ERROR = "Unknown protocol header b'$\\x11'"
        with open(os.path.join(DIRNAME, "pygpsdata-mixed.log"), "rb") as stream:
            with self.assertRaises(NMEAParseError) as context:
                i = 0
                raw = 0
                nmr = NMEAReader(stream, nmeaonly=True, quitonerror=ERR_RAISE)
                while raw is not None:
                    (raw, _) = nmr.read()
                    if raw is not None:
                        i += 1
        self.assertTrue(EXPECTED_ERROR in str(context.exception))

    def testNMEAMARITIME(self):  # test NMEA maritime messages
        EXPECTED_RESULTS = (
            "<NMEA(IIALF, numSen=1, senNum=1, seqmid=0, timelastchange=12:43:04.500000, alertcat=A, alertpriority=W, alertstate=A, mfrcode=, alertid=192, alertinst=1, revisionctr=1, escalationctr=0, alerttxt=LOST TARGET)>",
            "<NMEA(IIALF, numSen=2, senNum=1, seqmid=1, timelastchange=08:19:50.100000, alertcat=B, alertpriority=A, alertstate=S, mfrcode=XYZ, alertid=0512, alertinst=1, revisionctr=2, escalationctr=0, alerttxt=HEADING LOST)>",
            "<NMEA(IIALF, numSen=2, senNum=2, seqmid=1, timelastchange=, alertcat=, alertpriority=, alertstate=, mfrcode=XYZ, alertid=0512, alertinst=1, revisionctr=2, escalationctr=0, alerttxt=NO SYSTEM HEADING AVAILABLE)>",
            "<NMEA(AIEPV, status=R, equipmenttype=AI, equipmentid=503123450, propertyid=101, value=38400)>",
            "<NMEA(VRGEN, index=0000, time=01:12:00, data_01=0123, data_02=4567, data_03=89AB, data_04=CDEF, data_05=0123, data_06=4567, data_07=89AB, data_08=CDEF)>",
            "<NMEA(VRGEN, index=0008, time=01:12:00, data_01=0123, data_02=4567)>",
            "<NMEA(INNRM, function=2, freqtable=1, txcovermask=00001E1F, msgtypemask=00000023, status=C)>",
            "<NMEA(CRNRX, numSen=7, senNum=1, seqid=0, msgcode=IE69, freqtable=1, time=13:56:00, day=27, month=6, year=2001, totalchar=241.0, totalbad=3.0, status=A, body=============================)>",
            "<NMEA(CRNRX, numSen=7, senNum=2, seqid=0, msgcode=, freqtable=, time=, day=, month=, year=, totalchar=, totalbad=, status=, body=========^0D^0AISSUED ON SATURDAY 06 JANUARY 2001.)>",
            "<NMEA(CSSM3, MSIstatus=123456, msgnum=5213, LESseqnum=798, LESid=0, oceanregioncode=3, prioritycode=14, servicecode=0, presentationcode=2012, year=4, month=5, day=14, hour=30, minute=3400, circarealat=, circareaNS=076, circarealon=, circareaEW=300)>",
            "<NMEA(CSSMB, numSen=8, senNum=1, seqid=0, msgnum=123456, body=FROM:Maritime Rescue Coordination Centre xxx^0D)0ATO:)>",
            "<NMEA(IISPW, pwdprotectsentence=EPV, id=211000001, pwdlevel=2, password=SESAME)>",
            "<NMEA(IIEPV, status=C, equipmenttype=AI, equipmentid=211000001, propertyid=111, value=HEUREKA143)>",
            "<NMEA(AIEPV, status=R, equipmenttype=AI, equipmentid=211000001, propertyid=111, value=HEUREKA143)>",
        )

        i = 0
        raw = 0
        with open(os.path.join(DIRNAME, "pygpsdata-maritime.log"), "rb") as stream:
            nmr = NMEAReader(stream, nmeaonly=True, quitonerror=2)
            for raw, parsed in nmr:
                if raw is not None:
                    # print(f'"{parsed}",')
                    self.assertEqual(str(parsed), EXPECTED_RESULTS[i])
                    i += 1
        self.assertEqual(i, 14)

    def testNMEAITER(self):  # NMEAReader iterator
        EXPECTED_RESULTS = (
            "<NMEA(GNDTM, datum=W84, subDatum=, latOfset=0.0, NS=N, lonOfset=0.0, EW=E, alt=0.0, refDatum=W84)>",
            "<NMEA(GNRMC, time=10:36:07, status=A, lat=53.450657, NS=N, lon=-2.2404103333, EW=W, spd=0.046, cog=, date=2021-03-06, mv=, mvEW=, posMode=A, navStatus=V)>",
            "<NMEA(GNVTG, cogt=, cogtUnit=T, cogm=, cogmUnit=M, sogn=0.046, sognUnit=N, sogk=0.085, sogkUnit=K, posMode=A)>",
            "<NMEA(GNGNS, time=10:36:07, lat=53.450657, NS=N, lon=-2.2404103333, EW=W, posMode=AANN, numSV=6, HDOP=5.88, alt=56.0, sep=48.5, diffAge=, diffStation=, navStatus=V)>",
            "<NMEA(GNGGA, time=10:36:07, lat=53.450657, NS=N, lon=-2.2404103333, EW=W, quality=1, numSV=6, HDOP=5.88, alt=56.0, altUnit=M, sep=48.5, sepUnit=M, diffAge=, diffStation=)>",
            "<NMEA(GNGSA, opMode=A, navMode=3, svid_01=23, svid_02=24, svid_03=20, svid_04=12, svid_05=, svid_06=, svid_07=, svid_08=, svid_09=, svid_10=, svid_11=, svid_12=, PDOP=9.62, HDOP=5.88, VDOP=7.62, systemId=1)>",
            "<NMEA(GPGSV, numMsg=3, msgNum=1, numSV=11, svid_01=1, elv_01=6.0, az_01=14, cno_01=8, svid_02=12, elv_02=43.0, az_02=207, cno_02=28, svid_03=14, elv_03=6.0, az_03=49, cno_03=, svid_04=15, elv_04=44.0, az_04=171, cno_04=23, signalID=1)>",
            "<NMEA(GPTHS, headt=23.34, mi=A)>",
        )

        i = 0
        raw = 0
        with open(os.path.join(DIRNAME, "pygpsdata-nmea4sm.log"), "rb") as stream:
            nmr = NMEAReader(stream, nmeaonly=False)
            for raw, parsed in nmr:
                if raw is not None:
                    self.assertEqual(str(parsed), EXPECTED_RESULTS[i])
                    i += 1
        self.assertEqual(i, 8)

    def testNMEAITERATE(self):  # NMEAReader helper method
        EXPECTED_RESULTS = (
            "<NMEA(GNDTM, datum=W84, subDatum=, latOfset=0.0, NS=N, lonOfset=0.0, EW=E, alt=0.0, refDatum=W84)>",
            "<NMEA(GNRMC, time=10:36:07, status=A, lat=53.450657, NS=N, lon=-2.2404103333, EW=W, spd=0.046, cog=, date=2021-03-06, mv=, mvEW=, posMode=A, navStatus=V)>",
            "<NMEA(GNVTG, cogt=, cogtUnit=T, cogm=, cogmUnit=M, sogn=0.046, sognUnit=N, sogk=0.085, sogkUnit=K, posMode=A)>",
            "<NMEA(GNGNS, time=10:36:07, lat=53.450657, NS=N, lon=-2.2404103333, EW=W, posMode=AANN, numSV=6, HDOP=5.88, alt=56.0, sep=48.5, diffAge=, diffStation=, navStatus=V)>",
            "<NMEA(GNGGA, time=10:36:07, lat=53.450657, NS=N, lon=-2.2404103333, EW=W, quality=1, numSV=6, HDOP=5.88, alt=56.0, altUnit=M, sep=48.5, sepUnit=M, diffAge=, diffStation=)>",
            "<NMEA(GNGSA, opMode=A, navMode=3, svid_01=23, svid_02=24, svid_03=20, svid_04=12, svid_05=, svid_06=, svid_07=, svid_08=, svid_09=, svid_10=, svid_11=, svid_12=, PDOP=9.62, HDOP=5.88, VDOP=7.62, systemId=1)>",
            "<NMEA(GPGSV, numMsg=3, msgNum=1, numSV=11, svid_01=1, elv_01=6.0, az_01=14, cno_01=8, svid_02=12, elv_02=43.0, az_02=207, cno_02=28, svid_03=14, elv_03=6.0, az_03=49, cno_03=, svid_04=15, elv_04=44.0, az_04=171, cno_04=23, signalID=1)>",
            "<NMEA(GPTHS, headt=23.34, mi=A)>",
        )

        i = 0
        raw = 0
        with open(os.path.join(DIRNAME, "pygpsdata-nmea4sm.log"), "rb") as stream:
            nmr = NMEAReader(stream, nmeaonly=False)
            for raw, parsed in nmr:
                if raw is not None:
                    self.assertEqual(str(parsed), EXPECTED_RESULTS[i])
                    i += 1
        self.assertEqual(i, 8)

    def testNMEAITERATE_ERR1(
        self,
    ):  # NMEAReader iterator with bad checksum
        EXPECTED_ERROR = "Message GNVTG invalid checksum 3) - should be 30"
        with open(os.path.join(DIRNAME, "pygpsdata-nmeabadck2.log"), "rb") as stream:
            with self.assertRaises(NMEAParseError) as context:
                nmr = NMEAReader(
                    stream,
                    nmeaonly=False,
                    validate=VALCKSUM,
                    msgmode=0,
                    quitonerror=ERR_RAISE,
                )
                for raw, parsed in nmr:
                    pass
        self.assertTrue(EXPECTED_ERROR in str(context.exception))

    def testNMEAITERATE_ERR2(
        self,
    ):  # NMEAReader iterator ignoring bad checksum and passing error handler
        EXPECTED_RESULT = "<NMEA(GPGSV, numMsg=3, msgNum=1, numSV=11, svid_01=1, elv_01=0.0, az_01=32, cno_01=, svid_02=10, elv_02=27.0, az_02=310, cno_02=, svid_03=12, elv_03=19.0, az_03=205, cno_03=19, svid_04=13, elv_04=38.0, az_04=134, cno_04=21, signalID=1)>"
        with open(os.path.join(DIRNAME, "pygpsdata-nmeabadck2.log"), "rb") as stream:
            nmr = NMEAReader(
                stream,
                nmeaonly=False,
                validate=VALCKSUM,
                msgmode=0,
                quitonerror=ERR_LOG,
                errorhandler=lambda e: print(f"I ignored the following error: {e}"),
            )
            res = ""
            for raw, parsed in nmr:
                res = str(parsed)
            self.assertEqual(EXPECTED_RESULT, res)

    def testNMEAITERATE_ERR3(
        self,
    ):  # NMEAReader iterator ignoring bad checksum and continuing
        EXPECTED_RESULT = "<NMEA(GPGSV, numMsg=3, msgNum=1, numSV=11, svid_01=1, elv_01=0.0, az_01=32, cno_01=, svid_02=10, elv_02=27.0, az_02=310, cno_02=, svid_03=12, elv_03=19.0, az_03=205, cno_03=19, svid_04=13, elv_04=38.0, az_04=134, cno_04=21, signalID=1)>"
        with open(os.path.join(DIRNAME, "pygpsdata-nmeabadck2.log"), "rb") as stream:
            nmr = NMEAReader(
                stream,
                nmeaonly=False,
                validate=VALCKSUM,
                msgmode=0,
                quitonerror=ERR_IGNORE,
            )
            res = ""
            for raw, parsed in nmr:
                res = str(parsed)
            self.assertEqual(EXPECTED_RESULT, res)

    def testNMEAFOO1(self):  # stream containing invalid attribute type
        EXPECTED_ERROR = "Unknown attribute type Z2"
        with open(os.path.join(DIRNAME, "pygpsdata-nmeafoo1.log"), "rb") as stream:
            with self.assertRaises(NMEATypeError) as context:
                i = 0
                raw = 0
                nmr = NMEAReader(
                    stream,
                    nmeaonly=False,
                    quitonerror=ERR_RAISE,
                )
                for raw, parsed in nmr:
                    i += 1
        self.assertTrue(EXPECTED_ERROR in str(context.exception))

    def testNMEAFOO2(self):  # stream containing invalid value for attribute type
        EXPECTED_ERROR = "Incorrect type for attribute spd in msgID RMC"
        with open(os.path.join(DIRNAME, "pygpsdata-nmeafoo2.log"), "rb") as stream:
            with self.assertRaises(NMEATypeError) as context:
                i = 0
                raw = 0
                nmr = NMEAReader(stream, nmeaonly=False, quitonerror=ERR_RAISE)
                for raw, parsed in nmr:
                    i += 1
        self.assertTrue(EXPECTED_ERROR in str(context.exception))

    def testNMEABADMODE(self):  # invalid stream mode
        EXPECTED_ERROR = "Invalid stream mode 4 - must be 0, 1 or 2."
        with open(os.path.join(DIRNAME, "pygpsdata-nmeafoo1.log"), "rb") as stream:
            with self.assertRaises(NMEAParseError) as context:
                NMEAReader(stream, nmeaonly=False, validate=1, msgmode=4)
        self.assertTrue(EXPECTED_ERROR in str(context.exception))

    def testBADEOF(self):  # stream with premature EOF - should just be tolerated
        EXPECTED_RESULTS = (
            "<NMEA(GNDTM, datum=W84, subDatum=, latOfset=0.0, NS=N, lonOfset=0.0, EW=E, alt=0.0, refDatum=W84)>",
            "<NMEA(GNRMC, time=10:36:07, status=A, lat=53.450657, NS=N, lon=-2.2404103333, EW=W, spd=0.046, cog=, date=2021-03-06, mv=, mvEW=, posMode=A, navStatus=V)>",
            "<NMEA(GNVTG, cogt=, cogtUnit=T, cogm=, cogmUnit=M, sogn=0.046, sognUnit=N, sogk=0.085, sogkUnit=K, posMode=A)>",
            "<NMEA(GNGNS, time=10:36:07, lat=53.450657, NS=N, lon=-2.2404103333, EW=W, posMode=AANN, numSV=6, HDOP=5.88, alt=56.0, sep=48.5, diffAge=, diffStation=, navStatus=V)>",
        )

        i = 0
        raw = 0
        with open(os.path.join(DIRNAME, "pygpsdata-badeof.log"), "rb") as stream:
            nmr = NMEAReader(stream, nmeaonly=False)
            for raw, parsed in nmr:
                if raw is not None:
                    # print(parsed)
                    self.assertEqual(str(parsed), EXPECTED_RESULTS[i])
                    i += 1

    def testNMEATRIMBLE(self):  # test proprietary Trimble messages
        EXPECTED_RESULTS = (
            "<NMEA(GPLLQ, utctime=03:41:37, utcdate=2012-07-21, easting=23.45, eunit=M, northing=13.07, nunit=M, gpsQual=3, sip=15, posQual=0.011, height=3.14, hunit=M)>",
            "<NMEA(GPROT, rot=35.6, valid=A)>",
            "<NMEA(PFUGDP, type=GN, utctime=03:36:15, lat=39.8980003333, NS=N, lon=-105.112554, EW=W, siv=13, dpvoaQual=9, dgnssMode=FF, smajErr=0.1, sminErr=0.1, dirErr=149)>",
            "<NMEA(PTNLAVR, msgId=AVR, utctime=21:24:05.200000, yaw=52.1531, yawc=Yaw, tilt=-0.0806, tiltc=Tilt, roll=, rollc=, range=12.575, gpsQual=3, PDOP=1.4, sip=16)>",
            "<NMEA(PTNLAVR, msgId=AVR, utctime=21:26:04.300000, yaw=52.18, yawc=Yaw, tilt=, tiltc=, roll=-0.0807, rollc=Roll, range=12.579, gpsQual=3, PDOP=1.4, sip=16)>",
            "<NMEA(PTNLBPQ, msgId=BPQ, utctime=22:44:45.060000, utcdate=2007-12-02, lat=37.384897319, NS=N, lon=-122.0054366887, EW=W, height=EHT-5.923, hunit=M, gpsQual=5)>",
            "<NMEA(PTNLDG, msgId=DG, strength=44.0, snr=33.0, freq=287.0, bitRate=100, chan=0, trkStatus=4, trkPerf=1)>",
            "<NMEA(PTNLDG, msgId=DG, strength=124.0, snr=10.5, freq=1557855.0, bitRate=1200, chan=2, trkStatus=4, trkPerf=0)>",
            "<NMEA(PTNLEVT, msgId=EVT, utctime=22:12:12.000008, port=1, numEvents=5026, wno=1893, dow=1, leaps=17)>",
            "<NMEA(PTNLEVT, msgId=EVT, utctime=22:12:13.000008, port=1, numEvents=5027, wno=1893, dow=1, leaps=17)>",
            "<NMEA(PTNLGGK, msgId=GGK, utctime=10:29:39, utcdate=, lat=50.0162206402, NS=N, lon=8.4603351237, EW=E, gpsQual=5, sip=9, DOP=1.9, height=EHT150.790, hunit=M)>",
            "<NMEA(PTNLPJK, msgId=PJK, utctime=20:28:31.500000, utcdate=2012-11-01, northing=805083.35, nunit=N, easting=388997.346, eunit=E, gpsQual=10, sip=9, DOP=1.5, height=GHT+25.478, hunit=M)>",
            "<NMEA(PTNLPJK, msgId=PJK, utctime=01:07:17, utcdate=, northing=732646.511, nunit=N, easting=1731051.091, eunit=E, gpsQual=1, sip=5, DOP=2.7, height=EHT+28.345, hunit=M)>",
            "<NMEA(PTNLPJT, msgId=PJT, coordName=NAD83(Conus), projName=California Zone 4 0404)>",
            "<NMEA(PTNLVHD, msgId=VHD, utctime=03:05:56, utcdate=1998-09-30, azi=187.718, aziRate=-22.138, ele=-76.929, eleRate=-5.015, range=0.033, rangeRate=0.006, gpsQual=3, sip=7, PDOP=2.4, unit=M)>",
            "<NMEA(PTNLVGK, msgId=VGK, utctime=16:01:59, utcdate=1997-01-09, vectE=-0.161, vectN=9.985, vectV=-0.002, gpsQual=3, sip=7, DOP=1.0, vunit=4)>",
            "<NMEA(PASHRARR, msgId=ARR, vectNum=1, vectMode=3, sip=12, utctime=16:01:59, antEcefX=123.45, antEcefY=123.45, antEcefZ=-123.45, coord1std=12.34, coord2std=12.34, coord3std=12.34, coord12corr=2.34, coord13corr=2.34, coord23corr=-2.34, refId=S, vectFrame=0, vectOpt=2, clkAssum=1)>",
            "<NMEA(PASHRBTS, msgId=BTS, port_01=C, connected_01=1, name_01=btsdev1, addr_01=hs-344-fg, linkQual_01=87, port_02=H, connected_02=1, name_02=btsdev2, addr_02=pc-377xs, linkQual_02=68, port_03=T, connected_03=0, name_03=, addr_03=, linkQual_03=)>",
            "<NMEA(PGPPADV110, msgId=110, lat=39.88113582, lon=-105.07838455, height=1614.125)>",
            "<NMEA(PGPPADV120, msgId=120, prn_01=21, ele_01=76.82, azi_01=68.51, prn_02=29, ele_02=20.66, azi_02=317.47)>",
            "<NMEA(PASHR, utctime=14:25:09, trueHdg=179.885, trueHdgInd=T, roll=-0.624, pitch=0.245, reserved=, rollAcc=0.029, pitchAcc=0.029, hdgAcc=0.502, gnssQual=2, imuAlign=3)>",
        )

        i = 0
        raw = 0
        with open(os.path.join(DIRNAME, "trimble_nmea.log"), "rb") as stream:
            nmr = NMEAReader(stream, nmeaonly=False)
            for raw, parsed in nmr:
                if raw is not None:
                    # print(parsed)
                    self.assertEqual(str(parsed), EXPECTED_RESULTS[i])
                    i += 1
        self.assertEqual(i, 21)

    def testNMEAFURUNO(self):  # test proprietary Furuno messages
        EXPECTED_RESULTS = (
            "<NMEA(PFECGPatt, msgId=GPatt, yaw=12.345, pitch=23.456, roll=34.567)>",
            "<NMEA(PFECGPhve, msgId=GPhve, heave=12.345, status=A)>",
        )

        i = 0
        raw = 0
        with open(os.path.join(DIRNAME, "furuno_nmea.log"), "rb") as stream:
            nmr = NMEAReader(stream, nmeaonly=False, quitonerror=2)
            for raw, parsed in nmr:
                if raw is not None:
                    # print(f'"{parsed}",')
                    self.assertEqual(str(parsed), EXPECTED_RESULTS[i])
                    i += 1
        self.assertEqual(i, 2)

    def testNMEAKENWOOD(self):  # test proprietary Kenwood messages
        EXPECTED_RESULTS = (
            "<NMEA(PKWDWPL, time=15:08:03, status=V, lat=42.619, NS=N, lon=-71.3471666667, EW=W, sog=, cog=, date=2016-03-19, alt=, wpt=test, ts=5)>",
            "<NMEA(PKLDS, time=00:12:35, status=A, lat=35.7444166667, NS=N, lon=139.6698333333, EW=E, sog=15.0, cog=38.8, dat=1998-04-11, declination=10.8, dec_dir=W00, fleet=100.0, senderid=2000, senderstatus=15.0, reserved=0.0)>",
            "<NMEA(PKNDS, time=12:46:40, status=A, lat=49.90243, NS=N, lon=-119.39332, EW=W, sog=0.0, cog=0.0, date=2023-02-12, declination=19.2, dec_dir=W00, senderid=U00002, senderstatus=207.0, reserved=0.0)>",
            "<NMEA(PKLSH, lat=40.0, NS=N, lon=135.0, EW=E, time=02:17:20, status=A, fleetId=100, deviceId=2000)>",
            "<NMEA(PKNSH, lat=40.0, NS=N, lon=135.0, EW=E, time=02:17:20, status=A, senderid=U00001)>",
        )

        i = 0
        raw = 0
        with open(os.path.join(DIRNAME, "kenwood_nmea.log"), "rb") as stream:
            nmr = NMEAReader(stream, nmeaonly=False, quitonerror=2)
            for raw, parsed in nmr:
                if raw is not None:
                    # print(f'"{parsed}",')
                    self.assertEqual(str(parsed), EXPECTED_RESULTS[i])
                    i += 1
        self.assertEqual(i, 5)

    def testNMEALOCOSYSGET(self):  # test proprietary Locosys GET messages
        EXPECTED_RESULTS = (
            "<NMEA(PLSVD, velE=1.0, velN=0.0, velD=-16.0, velEdev=16.0, velNdev=15.0, velDdev=13.0)>",
            "<NMEA(PLSR, msgType=SLOPE, value=1, response=OK)>",
            "<NMEA(PLSR, msgType=SLOPE, value=0, response=OK)>",
            "<NMEA(PLSR, msgType=MEMS, value=1, response=OK)>",
            "<NMEA(PLSR, msgType=MEMS, value=0, response=OK)>",
            "<NMEA(PLSR, msgType=ATTIT, value=1, response=OK)>",
            "<NMEA(PLSR, msgType=ATTIT, value=0, response=OK)>",
            "<NMEA(PINVMSTR, value=0)>",
            "<NMEA(PINVMSLOPE, slope=-3.13, altDiff=-0.05, moveDist=0.93, slopeAccu=54.42, altDiffAccu=2.6, moveDistAccu=1.86)>",
            "<NMEA(PINVMIMU, timeSecond=1114.106, accelX=-0.3699, accelY=1.51074, accelZ=9.81383, gyroX=0.67139, gyroY=0.61035, gyroZ=-0.30518)>",
            "<NMEA(PINVMATTIT, roll=-20.652, pitch=32.265, yaw=0.0)>",
        )
        i = 0
        raw = 0
        with open(os.path.join(DIRNAME, "locosys_get_nmea.log"), "rb") as stream:
            nmr = NMEAReader(stream, nmeaonly=False, quitonerror=2)
            for raw, parsed in nmr:
                if raw is not None:
                    # print(f'"{parsed}",')
                    self.assertEqual(str(parsed), EXPECTED_RESULTS[i])
                    i += 1
        self.assertEqual(i, 11)

    def testNMEALOCOSYSSET(self):  # test proprietary Locosys SET messages
        EXPECTED_RESULTS = (
            "<NMEA(PINVCRES, value=0)>",
            "<NMEA(PINVCSTR, value=14)>",
            "<NMEA(PLSC, msgType=SLOPE, value=1)>",
            "<NMEA(PLSC, msgType=SLOPE, value=0)>",
            "<NMEA(PLSC, msgType=MEMS, value=1)>",
            "<NMEA(PLSC, msgType=MEMS, value=0)>",
            "<NMEA(PLSC, msgType=ATTIT, value=1)>",
            "<NMEA(PLSC, msgType=ATTIT, value=0)>",
            "<NMEA(PLSC, msgType=VER)>",
        )
        i = 0
        raw = 0
        with open(os.path.join(DIRNAME, "locosys_set_nmea.log"), "rb") as stream:
            nmr = NMEAReader(stream, nmeaonly=False, quitonerror=2, msgmode=SET)
            for raw, parsed in nmr:
                if raw is not None:
                    # print(f'"{parsed}",')
                    self.assertEqual(str(parsed), EXPECTED_RESULTS[i])
                    i += 1
        self.assertEqual(i, 9)

    def testNMEASEPX5GET(self):  # test proprietary Sepentrio X5 GET messages
        EXPECTED_RESULTS = (
            "<NMEA(PSSNHRP, msgId=HRP, utc=10:47:51, date=2024-03-23, hdg=23.455, roll=1.954, pitch=0.0125, hdgstd=0.123, rollstd=0.0234, pitchstd=0.03765, sip=11, attmode=0, magvar=4.56453, magvardir=W)>",
            "<NMEA(PSSNTFM, msgId=TFM, utc=10:47:51, heightind=2, usage2122=1021, usage2324=1023, usage252627=1025)>",
            "<NMEA(PSSNSNC, msgId=SNC, rev=0, tim=379359000, week=1840, cdidx_01=1, status_01=2, errcode_01=0, info_01=0)>",
            "<NMEA(PSSNSNC, msgId=SNC, rev=0, tim=379359123, week=1841, cdidx_01=1, status_01=2, errcode_01=0, info_01=0, cdidx_02=2, status_02=3, errcode_02=0, info_02=0)>",
        )
        i = 0
        raw = 0
        with open(os.path.join(DIRNAME, "septentriox5_nmea.log"), "rb") as stream:
            nmr = NMEAReader(stream, nmeaonly=False, quitonerror=2)
            for raw, parsed in nmr:
                if raw is not None:
                    # print(f'"{parsed}",')
                    self.assertEqual(str(parsed), EXPECTED_RESULTS[i])
                    i += 1
        self.assertEqual(i, 4)

    def testNMEAGPFMI(self):  # test proprietary Feyman GPFMI IMU GET messages

        def getstatus(status) -> str:

            st = ""
            for key, (desc, _) in FMI_STATUS.items():
                if int(status, 16) & key:
                    st += f"{desc} "
            return st

        EXPECTED_RESULTS = (
            "<NMEA(GPFMI, time=16:02:24.400000, lat=0.0, lon=0.0, alt=0.0, roll=3.0182, pitch=0.0573, yaw=9.99999, hrms=0.17, numSV=12, diffAge=0, navmod=2, status=001C0001)>",
            "<NMEA(GPFMI, time=16:02:24.600000, lat=0.0, lon=0.0, alt=0.0, roll=3.0182, pitch=0.0573, yaw=9.99999, hrms=0.17, numSV=12, diffAge=0, navmod=2, status=001C0001)>",
            "<NMEA(GPFMI, time=21:39:37.460000, lat=13.423023, lon=113.193168, alt=45.890494, roll=16.593, pitch=-7.75, yaw=3.888, hrms=0.17, numSV=26, diffAge=5, navmod=2, status=001C0001)>",
            "<NMEA(GPFMI, time=21:39:37.460000, lat=13.512547, lon=113.637276, alt=45.767895, roll=-3.523, pitch=-7.218, yaw=9.242, hrms=0.17, numSV=27, diffAge=5, navmod=2, status=001C0001)>",
            "<NMEA(GPFMI, time=21:39:37.460000, lat=13.19525, lon=113.882219, alt=45.719773, roll=4.998, pitch=-8.686, yaw=0.927, hrms=0.17, numSV=27, diffAge=3, navmod=2, status=001C0001)>",
            "<NMEA(GPFMI, time=21:39:37.460000, lat=13.092634, lon=113.283257, alt=45.018596, roll=8.221, pitch=4.892, yaw=-5.005, hrms=0.17, numSV=26, diffAge=4, navmod=2, status=001C0001)>",
            "<NMEA(GPFMI, time=21:39:37.460000, lat=13.254565, lon=113.288853, alt=45.661074, roll=7.074, pitch=3.049, yaw=-4.613, hrms=0.17, numSV=25, diffAge=3, navmod=2, status=001C0001)>",
            "<NMEA(GPFMI, time=21:39:37.460000, lat=13.509901, lon=113.138161, alt=45.996202, roll=-3.514, pitch=-5.205, yaw=8.724, hrms=0.17, numSV=27, diffAge=3, navmod=2, status=001C0001)>",
            "<NMEA(GPFMI, time=21:39:37.460000, lat=13.76182, lon=113.482807, alt=45.764244, roll=-4.72, pitch=6.391, yaw=-8.178, hrms=0.17, numSV=26, diffAge=2, navmod=2, status=001C0001)>",
            "<NMEA(GPFMI, time=21:39:37.460000, lat=13.934536, lon=113.034356, alt=45.252132, roll=16.878, pitch=2.358, yaw=-2.944, hrms=0.17, numSV=27, diffAge=2, navmod=2, status=001C0001)>",
            "<NMEA(GPFMI, time=21:39:37.460000, lat=13.639909, lon=113.402712, alt=45.615865, roll=11.508, pitch=9.446, yaw=3.934, hrms=0.17, numSV=27, diffAge=5, navmod=2, status=001C0001)>",
            "<NMEA(GPFMI, time=21:39:37.460000, lat=13.269057, lon=113.727968, alt=45.891641, roll=-0.085, pitch=6.887, yaw=-16.113, hrms=0.17, numSV=25, diffAge=2, navmod=2, status=001C0001)>",
            "<NMEA(GPFMI, time=21:39:37.460000, lat=13.457734, lon=113.13478, alt=45.028494, roll=6.838, pitch=2.27, yaw=8.636, hrms=0.17, numSV=24, diffAge=4, navmod=2, status=001C0001)>",
            "<NMEA(GPFMI, time=21:39:37.460000, lat=13.114631, lon=113.083719, alt=45.146318, roll=6.506, pitch=7.586, yaw=-3.084, hrms=0.17, numSV=26, diffAge=4, navmod=2, status=001C0001)>",
            "<NMEA(GPFMI, time=21:39:37.460000, lat=13.561758, lon=113.504337, alt=45.81872, roll=-19.365, pitch=1.583, yaw=-7.598, hrms=0.17, numSV=25, diffAge=3, navmod=2, status=001C0001)>",
            "<NMEA(GPFMI, time=21:39:37.460000, lat=13.484163, lon=113.286151, alt=45.117093, roll=-7.733, pitch=3.219, yaw=-12.475, hrms=0.17, numSV=24, diffAge=3, navmod=2, status=001C0001)>",
            "<NMEA(GPFMI, time=21:39:37.460000, lat=13.077227, lon=113.853113, alt=45.633862, roll=16.283, pitch=-9.75, yaw=12.718, hrms=0.17, numSV=24, diffAge=5, navmod=2, status=001C0001)>",
            "<NMEA(GPFMI, time=21:39:37.460000, lat=13.256628, lon=113.893556, alt=45.736646, roll=-13.055, pitch=-6.64, yaw=5.213, hrms=0.17, numSV=26, diffAge=2, navmod=2, status=001C00E1)>",
            "<NMEA(GPFMI, time=21:39:37.460000, lat=13.327387, lon=113.710465, alt=45.119112, roll=-15.582, pitch=0.837, yaw=-8.4, hrms=0.17, numSV=24, diffAge=2, navmod=2, status=001C0001)>",
            "<NMEA(GPFMI, time=21:39:37.460000, lat=13.238952, lon=113.301512, alt=45.781582, roll=0.25, pitch=6.982, yaw=-3.916, hrms=0.17, numSV=26, diffAge=4, navmod=2, status=001C0001)>",
            "<NMEA(GPFMI, time=21:39:37.460000, lat=13.8632, lon=113.998859, alt=45.163301, roll=5.528, pitch=-2.215, yaw=16.083, hrms=0.17, numSV=24, diffAge=2, navmod=2, status=001A0001)>",
            "<NMEA(GPFMI, time=21:39:37.460000, lat=13.952284, lon=113.441564, alt=45.696884, roll=-12.544, pitch=-3.978, yaw=-2.798, hrms=0.17, numSV=25, diffAge=2, navmod=2, status=001A0001)>",
        )
        EXPECTED_RESULTS2 = (
            "Finit PPSReady SyncReady GnssConnect ",
            "Finit PPSReady SyncReady GnssConnect ",
            "Finit PPSReady SyncReady GnssConnect ",
            "Finit PPSReady SyncReady GnssConnect ",
            "Finit PPSReady SyncReady GnssConnect ",
            "Finit PPSReady SyncReady GnssConnect ",
            "Finit PPSReady SyncReady GnssConnect ",
            "Finit PPSReady SyncReady GnssConnect ",
            "Finit PPSReady SyncReady GnssConnect ",
            "Finit PPSReady SyncReady GnssConnect ",
            "Finit PPSReady SyncReady GnssConnect ",
            "Finit PPSReady SyncReady GnssConnect ",
            "Finit PPSReady SyncReady GnssConnect ",
            "Finit PPSReady SyncReady GnssConnect ",
            "Finit PPSReady SyncReady GnssConnect ",
            "Finit PPSReady SyncReady GnssConnect ",
            "Finit PPSReady SyncReady GnssConnect ",
            "Finit FReset FixRlsStage1 FixRlsStage2 PPSReady SyncReady GnssConnect ",
            "Finit PPSReady SyncReady GnssConnect ",
            "Finit PPSReady SyncReady GnssConnect ",
            "Finit FInitOk SyncReady GnssConnect ",
            "Finit FInitOk SyncReady GnssConnect ",
        )
        i = 0
        raw = 0
        with open(os.path.join(DIRNAME, "feyman_nmea.log"), "rb") as stream:
            nmr = NMEAReader(stream, nmeaonly=False, quitonerror=2)
            for raw, parsed in nmr:
                if raw is not None:
                    # print(f'"{parsed}",')
                    self.assertEqual(str(parsed), EXPECTED_RESULTS[i])
                    self.assertEqual(getstatus(parsed.status), EXPECTED_RESULTS2[i])
                    i += 1
        self.assertEqual(i, 22)

    def testBADHDR_FAIL(self):  # invalid header in data with quitonerror = 2
        EXPECTED_ERROR = "Unknown protocol header b'$&'."
        with self.assertRaises(NMEAParseError) as context:
            i = 0
            with open(os.path.join(DIRNAME, "pygpsdata-BADHDR.log"), "rb") as stream:
                ubr = NMEAReader(stream, quitonerror=ERR_RAISE, nmeaonly=True)
                for _, _ in ubr:
                    i += 1
        self.assertTrue(EXPECTED_ERROR in str(context.exception))

    def testBADHDR_LOG(self):  # invalid header in data with quitonerror = 1
        i = 0
        with self.assertLogs(level=ERROR) as log:
            with open(os.path.join(DIRNAME, "pygpsdata-BADHDR.log"), "rb") as stream:
                ubr = NMEAReader(stream, quitonerror=ERR_LOG, nmeaonly=True)
                for raw, parsed in ubr:
                    i += 1
            self.assertEqual(
                ["ERROR:pynmeagps.nmeareader:Unknown protocol header b'$&'."],
                log.output,
            )

    def testBADHDR_IGNORE(self):  # invalid header in data with quitonerror = 0
        i = 0
        with open(os.path.join(DIRNAME, "pygpsdata-BADHDR.log"), "rb") as stream:
            ubr = NMEAReader(stream, quitonerror=ERR_IGNORE, nmeaonly=True)
            for raw, parsed in ubr:
                i += 1
            self.assertEqual(i, 15)

    def testLCCHECKSUM(self):  # lower case checksum bytes
        EXPECTED_RESULTS = (
            "<NMEA(GNRMC, time=10:36:07, status=A, lat=53.450657, NS=N, lon=-102.2404103333, EW=W, spd=0.046, cog=, date=2021-03-06, mv=, mvEW=, posMode=A, navStatus=V)>",
            "<NMEA(GPGSV, numMsg=3, msgNum=1, numSV=11, svid_01=1, elv_01=6.0, az_01=14, cno_01=8, svid_02=12, elv_02=43.0, az_02=207, cno_02=28, svid_03=14, elv_03=6.0, az_03=49, cno_03=, svid_04=15, elv_04=44.0, az_04=171, cno_04=23, signalID=1)>",
            "<NMEA(GNZDA, time=10:36:07, day=6, month=3, year=2021, ltzh=00, ltzn=00)>",
            "<NMEA(GNGBS, time=10:36:07, errLat=15.1, errLon=24.2, errAlt=31.0, svid=, prob=, bias=, stddev=, systemId=, signalId=)>",
            "<NMEA(GNVLW, twd=, twdUnit=N, wd=, wdUnit=N, tgd=0.0, tgdUnit=N, gd=0.0, gdUnit=N)>",
        )
        i = 0
        with open(os.path.join(DIRNAME, "pygpsdata-nmea4lc.log"), "rb") as stream:
            ubr = NMEAReader(stream, quitonerror=ERR_RAISE)
            for raw, parsed in ubr:
                # print(f'"{parsed}",')
                self.assertEqual(str(parsed), EXPECTED_RESULTS[i])
                i += 1
            self.assertEqual(i, 5)

    def testuserdefined(self):  # test user-defined payload definition dictionary
        EXPECTED_RESULTS = (
            "<NMEA(U1XYZ, fixutc=10:29:29, lat=44.4506706667, NS=N, lon=-1.24026, EW=W)>",
            "<NMEA(U2ZYX, fixutc=10:38:29, data1=somedata, data2=somemoredata)>",
        )
        USERDEFINED = {
            "XYZ": {"fixutc": TM, "lat": LA, "NS": LAD, "lon": LN, "EW": LND},
            "ZYX": {"fixutc": TM, "data1": ST, "data2": ST},
        }
        i = 0
        with open(os.path.join(DIRNAME, "pygpsdata-userdefined.log"), "rb") as stream:
            ubr = NMEAReader(
                stream, quitonerror=ERR_RAISE, nmeaonly=True, userdefined=USERDEFINED
            )
            for raw, parsed in ubr:
                # print(f'"{parsed}",')
                self.assertEqual(str(parsed), EXPECTED_RESULTS[i])
                i += 1
            self.assertEqual(i, 2)


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



"""
socketwrapper.py

Socket stream wrapper providing read(n) and readline() methods.

Supports chunked and compressed transfer-encoded datastreams.

Created on 12 Feb 2023

:author: semuadmin (Steve Smith)
:copyright: semuadmin © 2023
:license: BSD 3-Clause
"""

import socket
from io import BytesIO
from logging import getLogger
from zlib import MAX_WBITS, decompress
from zlib import error as zlibError

from pynmeagps.nmeatypes_core import (
    DEFAULT_BUFSIZE,
    ENCODE_CHUNKED,
    ENCODE_COMPRESS,
    ENCODE_DEFLATE,
    ENCODE_GZIP,
    ENCODE_NONE,
)


class SocketWrapper:
    """
    Socket stream wrapper providing read(n) and readline() methods.

    Supports chunked transfer-encoded datastreams.
    """

    def __init__(self, sock: socket, encoding=ENCODE_NONE, bufsize=DEFAULT_BUFSIZE):
        """
        Constructor.

        :param sock socket: socket object
        :param int encoding: transfer-encoding values \
            (0 = none, 1 = chunk, 2 = gzip, 4 = compress, 8 = deflate (can be OR'd) (0)
        :param int bufsize: internal buffer size
        """

        # configure logger with name "pygnssutils" in calling module
        self.logger = getLogger(__name__)
        self._socket = sock
        self._bufsize = bufsize
        self._encoding = encoding
        self._buffer = bytearray()
        self._partial = b""  # partial chunk
        self._recv()  # populate initial buffer

    def _recv(self) -> bool:
        """
        Read bytes from socket into internal buffer.

        :returns: return code (0 = failure, 1 = success)
        :rtype: bool
        """

        try:
            data = self._socket.recv(self._bufsize)
            if len(data) == 0:
                return False
            if self._encoding & ENCODE_CHUNKED:
                data = self._partial + data
                chunks, self._partial = self.dechunk(data)
                self._buffer += chunks
            else:
                self._buffer += data
        except (OSError, TimeoutError):
            return False
        return True

    def read(self, num: int) -> bytes:
        """
        Read specified number of bytes from buffer.
        NB: always check length of return data.

        :param int num: number of bytes to read
        :returns: bytes read (which may be less than num)
        :rtype: bytes
        """

        # if at end of internal buffer, top it up from socket
        while len(self._buffer) < num:
            if not self._recv():
                return b""
        data = self._buffer[:num]
        self._buffer = self._buffer[num:]
        return bytes(data)

    def readline(self) -> bytes:
        """
        Read bytes from buffer until CRLF reached.
        NB: always check that return data terminator is CRLF.

        :returns: bytes
        :rtype: bytes
        """

        line = b""
        while True:
            data = self.read(1)
            if len(data) == 1:
                line += data
                if line[-2:] == b"\r\n":
                    break
            else:
                break

        return line

    def write(self, data: bytes, **kwargs):
        """
        Write bytes to socket.

        :param bytes data: data
        :param dict kwargs: kwargs
        """

        return self._socket.send(data, **kwargs)

    def in_waiting(self) -> int:
        """
        Return number of bytes in buffer.

        :returns: length of buffer
        :rtype: int
        """

        return len(self._buffer)

    def dechunk(self, segment: bytes) -> tuple:  # pragma: no cover
        """
        Parse segment of chunked transfer-encoded byte stream.

        Returns complete chunks in this segment and any partial
        chunk, which should be prepended to next segment read.

        :param segment: segment of byte stream
        :returns: tuple of (chunks, partial)
        :rtype: tuple
        """

        # pylint: disable=logging-fstring-interpolation

        instream = BytesIO(segment)
        chunks = b""
        partial = b""

        while True:
            length_bytes = instream.readline()
            if length_bytes[-2:] != b"\r\n":
                # premature end of length bytes
                partial = length_bytes
                break
            try:
                chunk_length = int(length_bytes.strip(), 16)
            except ValueError:
                # residual bytes at beginning of stream
                break
            if chunk_length != 0:
                chunk = instream.read(chunk_length)
                if len(chunk) != chunk_length:
                    # premature end of chunk bytes
                    partial = length_bytes + chunk
                    break
                try:
                    if self._encoding & ENCODE_GZIP:
                        chunk = decompress(chunk, wbits=MAX_WBITS | 16)
                    if self._encoding & ENCODE_COMPRESS:
                        chunk = decompress(chunk, wbits=MAX_WBITS)
                    if self._encoding & ENCODE_DEFLATE:
                        chunk = decompress(chunk, wbits=-MAX_WBITS)
                except zlibError as err:  # pragma: no cover
                    self.logger.error(f"Error decompressing data: {err}")
                    # parser will discard data
                chunks += chunk

            instream.readline()
            if chunk_length == 0:
                # final chunk
                break

        return chunks, partial

    @property
    def buffer(self) -> bytearray:
        """
        Getter for buffer.

        :return: buffer
        :rtype: bytearray
        """

        return self._buffer
